import importlib
import sys
from collections.abc import Callable
import logging 

import timm
import torch
from torch import nn

import terratorch.models.decoders as decoder_registry
from terratorch.models.backbones.clay_v1.embedder import Embedder
from terratorch.models.model import (
    AuxiliaryHead,
    AuxiliaryHeadWithDecoderWithoutInstantiatedHead,
    Model,
    ModelFactory,
)
from terratorch.models.pixel_wise_model import PixelWiseModel
from terratorch.models.scalar_output_model import ScalarOutputModel
from terratorch.models.utils import DecoderNotFoundError, extract_prefix_keys
from terratorch.registry import MODEL_FACTORY_REGISTRY

PIXEL_WISE_TASKS = ["segmentation", "regression"]
SCALAR_TASKS = ["classification", "scalar_regression"]
SUPPORTED_TASKS = PIXEL_WISE_TASKS + SCALAR_TASKS


class DecoderNotFoundError(Exception):
    pass

class ModelWrapper(nn.Module):

    def __init__(self, model: nn.Module = None) -> None:

        super(ModelWrapper, self).__init__()

        self.model = model

        self.embedding_shape = self.model.model.state_dict()['decoder.embed_to_pixels.dem.bias'].shape[0]

    def channels(self):
        return (1, self.embedding_shape)

    @property
    def parameters(self):
        return self.model.parameters

    def forward(self, args, **kwargs):
        datacube = {}
        datacube['pixels'] = args
        datacube['timestep'] = None
        datacube['latlon'] = None
        return self.model.forward(datacube)

@MODEL_FACTORY_REGISTRY.register
class ClayModelFactory(ModelFactory):
    def build_model(
        self,
        task: str,
        backbone: str | nn.Module,
        decoder: str | nn.Module,
        in_channels: int,
        bands: list[int] = [],
        num_classes: int | None = None,
        pretrained: bool = True,  # noqa: FBT001, FBT002
        num_frames: int = 1,
        prepare_features_for_image_model: Callable | None = None,
        aux_decoders: list[AuxiliaryHead] | None = None,
        rescale: bool = True,  # noqa: FBT002, FBT001
        checkpoint_path: str = None,
        **kwargs,
    ) -> Model:
        """Model factory for Clay models.

        Further arguments to be passed to the backbone, decoder or head. They should be prefixed with
        `backbone_`, `decoder_` and `head_` respectively.

        Args:
            task (str): Task to be performed. Currently supports "segmentation" and "regression".
            backbone (str, nn.Module): Backbone to be used. If string, should be able to be parsed
                by the specified factory. Defaults to "prithvi_100".
            decoder (Union[str, nn.Module], optional): Decoder to be used for the segmentation model.
                    If a string, it will be created from a class exposed in decoder.__init__.py with the same name.
                    If an nn.Module, we expect it to expose a property `decoder.out_channels`.
                    Will be concatenated with a Conv2d for the final convolution. Defaults to "FCNDecoder".
            in_channels (int, optional): Number of input channels. Defaults to 3.
            num_classes (int, optional): Number of classes. None for regression tasks.
            pretrained (Union[bool, Path], optional): Whether to load pretrained weights for the backbone, if available.
                Defaults to True.
            num_frames (int, optional): Number of timesteps for the model to handle. Defaults to 1.
            prepare_features_for_image_model (Callable | None): Function to be called on encoder features
                before passing them to the decoder. Defaults to None, which applies the identity function.
            aux_decoders (list[AuxiliaryHead] | None): List of AuxiliaryHead deciders to be added to the model.
                These decoders take the input from the encoder as well.
            rescale (bool): Whether to apply bilinear interpolation to rescale the model output if its size
                is different from the ground truth. Only applicable to pixel wise models (e.g. segmentation, pixel wise regression). Defaults to True.

        Raises:
            NotImplementedError: _description_
            DecoderNotFoundException: _description_

        Returns:
            nn.Module: _description_
        """
        if not torch.cuda.is_available():
            self.CPU_ONLY = True
        else:
            self.CPU_ONLY = False

        # Path for accessing the model source code.
        self.syspath_kwarg = "model_sys_path"
        backbone_kwargs, kwargs = extract_prefix_keys(kwargs, "backbone_")

        # TODO: support auxiliary heads
        if not isinstance(backbone, nn.Module):
            if not "clay" in backbone:
                msg = "This class only handles models for `Clay` encoders"
                raise NotImplementedError(msg)

            task = task.lower()
            if task not in SUPPORTED_TASKS:
                msg = f"Task {task} not supported. Please choose one of {SUPPORTED_TASKS}"
                raise NotImplementedError(msg)

            # Trying to find the model on HuggingFace.
            try:
                backbone: nn.Module = timm.create_model(
                    backbone,
                    pretrained=pretrained,
                    in_chans=in_channels,
                    bands=bands,
                    num_frames=num_frames,
                    features_only=True,
                    **backbone_kwargs,
                )
            except Exception as e:
                print(e, "Error loading from HF. Trying to instantiate locally ...")

        else:
            if checkpoint_path is None:
                raise ValueError("A checkpoint (checkpoint_path) must be provided to restore the model.")

            backbone: nn.Module = Embedder(ckpt_path=checkpoint_path, **backbone_kwargs)
            print("Model Clay was successfully restored.")

        # If patch size is not provided in the config or by the model, it might lead to errors due to irregular images.
        patch_size = backbone_kwargs.get("patch_size", None)
        if patch_size is None:
            # Infer patch size from model by checking all backbone modules
            for module in backbone.modules():
                if hasattr(module, "patch_size"):
                    patch_size = module.patch_size
                    break
        padding = backbone_kwargs.get("padding", "reflect")

        # allow decoder to be a module passed directly
        decoder_cls = _get_decoder(decoder)
        decoder_kwargs, kwargs = extract_prefix_keys(kwargs, "decoder_")

        # TODO: remove this
        decoder: nn.Module = decoder_cls(
            backbone.feature_info.channels(), **decoder_kwargs)
        # decoder: nn.Module = decoder_cls([128, 256, 512, 1024], **decoder_kwargs)

        head_kwargs, kwargs = extract_prefix_keys(kwargs, "head_")
        if num_classes:
            head_kwargs["num_classes"] = num_classes
        if aux_decoders is None:
            return _build_appropriate_model(
                task, backbone, decoder, head_kwargs, prepare_features_for_image_model, patch_size=patch_size, padding=padding, rescale=rescale
            )

        to_be_aux_decoders: list[AuxiliaryHeadWithDecoderWithoutInstantiatedHead] = []

        for aux_decoder in aux_decoders:
            args = aux_decoder.decoder_args if aux_decoder.decoder_args else {}
            aux_decoder_cls: nn.Module = _get_decoder(aux_decoder.decoder)
              
            aux_decoder_kwargs, kwargs = extract_prefix_keys(args, "decoder_")
            aux_decoder_instance = aux_decoder_cls(backbone.feature_info.channels(), **aux_decoder_kwargs)
            # aux_decoder_instance = aux_decoder_cls([128, 256, 512, 1024], **decoder_kwargs)

            aux_head_kwargs, kwargs = extract_prefix_keys(args, "head_")
            if num_classes:
                aux_head_kwargs["num_classes"] = num_classes
            # aux_head: nn.Module = _get_head(task, aux_decoder_instance, num_classes=num_classes, **head_kwargs)
            # aux_decoder.decoder = nn.Sequential(aux_decoder_instance, aux_head)
            to_be_aux_decoders.append(
                AuxiliaryHeadWithDecoderWithoutInstantiatedHead(
                    aux_decoder.name, aux_decoder_instance, aux_head_kwargs)
            )

        return _build_appropriate_model(
            task,
            backbone,
            decoder,
            head_kwargs,
            prepare_features_for_image_model,
            patch_size=patch_size,
            padding=padding,
            rescale=rescale,
            auxiliary_heads=to_be_aux_decoders,
        )


def _build_appropriate_model(
    task: str,
    backbone: nn.Module,
    decoder: nn.Module,
    head_kwargs: dict,
    prepare_features_for_image_model: Callable,
    patch_size: int | list | None,
    padding: str,
    rescale: bool = True,  # noqa: FBT001, FBT002
    auxiliary_heads: dict | None = None,
):
    if task in PIXEL_WISE_TASKS:
        return PixelWiseModel(
            task,
            backbone,
            decoder,
            head_kwargs,
            patch_size=patch_size,
            padding=padding,
            rescale=rescale,
            auxiliary_heads=auxiliary_heads,
        )
    elif task in SCALAR_TASKS:
        return ScalarOutputModel(
            task,
            backbone,
            decoder,
            head_kwargs,
            patch_size=patch_size,
            padding=padding,
            auxiliary_heads=auxiliary_heads,
        )


def _get_decoder(decoder: str | nn.Module) -> nn.Module:
    if isinstance(decoder, nn.Module):
        return decoder
    if isinstance(decoder, str):
        try:
            decoder = getattr(decoder_registry, decoder)
            return decoder
        except AttributeError as decoder_not_found_exception:
            msg = f"Decoder {decoder} was not found in the registry."
            raise DecoderNotFoundError(msg) from decoder_not_found_exception
    msg = "Decoder must be str or nn.Module"
    raise Exception(msg)



import pytest

from terratorch.registry import Registry
from terratorch.utils import InvalidModelError


class DummyModel:
    def __init__(self, param):
        self.param = param


@pytest.fixture
def registry():
    return Registry()


def test_register_function(registry):
    """Test if a function can be registered and retrieved."""

    @registry.register
    def dummy_function():
        return "dummy"

    assert "dummy_function" in registry


def test_register_class(registry):
    """Test if a class can be registered and instantiated."""

    @registry.register
    class DummyClass:
        def __init__(self, value):
            self.value = value

    assert "DummyClass" in registry
    instance = registry.build("DummyClass", value=5)
    assert isinstance(instance, DummyClass)
    assert instance.value == 5


def test_build_model(registry):
    """Test building a model with arguments."""

    @registry.register
    def build_dummy_model(param):
        return DummyModel(param)

    model = registry.build("build_dummy_model", param=42)
    assert isinstance(model, DummyModel)
    assert model.param == 42


def test_build_nonexistent_model(registry):
    """Test that an error is raised when trying to build an unregistered model."""
    with pytest.raises(InvalidModelError, match="build_nonexistent_model"):
        registry.build("build_nonexistent_model")


def test_len_and_iter(registry):
    """Test the length and iteration over the registry."""

    @registry.register
    def model_a():
        pass

    @registry.register
    def model_b():
        pass

    assert len(registry) == 2
    assert set(registry) == {"model_a", "model_b"}


def test_invalid_registration(registry):
    """Test that registering a non-callable object raises a TypeError."""
    with pytest.raises(TypeError):
        registry.register("not_a_function_or_class")


def test_contains_method(registry):
    """Test if the __contains__ method works as expected."""

    @registry.register
    def model_x():
        pass

    assert "model_x" in registry
    assert "nonexistent_model" not in registry



# Copyright contributors to the Terratorch project

from torch import Tensor, nn

from terratorch.registry import TERRATORCH_DECODER_REGISTRY


class Norm2d(nn.Module):
    def __init__(self, embed_dim):
        super().__init__()
        self.ln = nn.LayerNorm(embed_dim, eps=1e-6)

    def forward(self, x: Tensor):
        x = x.permute(0, 2, 3, 1)
        x = self.ln(x)
        x = x.permute(0, 3, 1, 2).contiguous()
        return x


def _conv_upscale_block(input_channels, output_channels, kernel_size, stride, dilation, padding, output_padding):
    return nn.Sequential(
        nn.ConvTranspose2d(
            input_channels,
            output_channels,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            padding=padding,
            output_padding=output_padding,
        ),
        nn.Conv2d(output_channels, output_channels, kernel_size=3, padding=1, bias=True),
        Norm2d(output_channels),
        nn.GELU(),
    )

@TERRATORCH_DECODER_REGISTRY.register
class FCNDecoder(nn.Module):
    """Fully Convolutional Decoder"""

    def __init__(self, embed_dim: int, channels: int = 256, num_convs: int = 4, in_index: int = -1) -> None:
        """Constructor

        Args:
            embed_dim (_type_): Input embedding dimension
            channels (int, optional): Number of channels for each conv. Defaults to 256.
            num_convs (int, optional): Number of convs. Defaults to 4.
            in_index (int, optional): Index of the input list to take. Defaults to -1.
        """
        super().__init__()
        kernel_size = 2
        stride = 2
        dilation = 1
        padding = 0
        output_padding = 0
        self.channels = channels
        self.num_convs = num_convs
        self.in_index = in_index
        self.embed_dim = embed_dim[in_index]
        if num_convs < 1:
            msg = "num_convs must be >= 1"
            raise Exception(msg)

        convs = []

        for i in range(num_convs):
            in_channels = self.embed_dim if i == 0 else self.channels
            convs.append(
                _conv_upscale_block(in_channels, self.channels, kernel_size, stride, dilation, padding, output_padding)
            )

        self.convs = nn.Sequential(*convs)

    @property
    def out_channels(self):
        return self.channels

    def forward(self, x: list[Tensor]):
        x = x[self.in_index]
        decoded = self.convs(x)
        return decoded



import pytest
import torch
from terratorch.models.backbones.clay_v1.modules import (
    FeedForward,
    Attention,
    Transformer,
    Encoder,
    EmbeddingEncoder,
    DynamicEmbedding,
    WavesTransformer,
    Datacuber,
    WAVELENGTHS
)

# ----------- FeedForward -----------
def test_feedforward_forward():
    x = torch.randn(2, 10, 16)
    ff = FeedForward(dim=16, hidden_dim=32)
    out = ff(x)
    assert out.shape == x.shape

# ----------- Attention -----------
def test_attention_forward(monkeypatch):
    x = torch.randn(2, 5, 16)
    attn = Attention(dim=16, heads=2, dim_head=8)

    # patch use_fused_attn to False to hit else branch
    monkeypatch.setattr("timm.layers.use_fused_attn", lambda: False)
    out = attn(x)
    assert out.shape == x.shape

# ----------- Transformer -----------
def test_transformer_forward():
    x = torch.randn(2, 5, 16)
    trans = Transformer(dim=16, depth=2, heads=2, dim_head=8, mlp_dim=32)
    out = trans(x)
    # out is list of tensor per layer
    assert isinstance(out, list)
    assert out[-1].shape == x.shape

def test_transformer_vpt_forward():
    x = torch.randn(2, 5, 16)
    trans = Transformer(dim=16, depth=2, heads=2, dim_head=8, mlp_dim=32, vpt=True, vpt_n_tokens=3)
    out = trans(x)

# ----------- Encoder -----------
def get_datacube(batch=2, C=3, H=4, W=4):
    cube = torch.randn(batch, C, H, W)
    time = torch.randn(batch, 2)
    latlon = torch.randn(batch, 2)
    gsd = torch.tensor(1.0)
    # waves should be wavelength values (1D), not pre-encoded features
    waves = torch.randn(C)
    return {"pixels": cube, "time": time, "latlon": latlon, "gsd": gsd, "waves": waves}

@pytest.mark.xfail(reason="Clay v1 Encoder has dimension mismatch issue in add_encodings - needs investigation")
def test_encoder_forward():
    """Test that Encoder forward pass works"""
    C = 6  # Use a reasonable number of channels
    datacube = get_datacube(batch=2, C=C, H=16, W=16)  # Larger image to work with standard patch sizes

    enc = Encoder(
        mask_ratio=0.5,
        patch_size=4,
        shuffle=True,
        dim=768,  # Standard Clay dimension
        depth=1,
        heads=12,
        dim_head=64,
        mlp_ratio=4.0,
    )

    # datacube already contains waves, no need to pass separately
    out, unmasked_idx, masked_idx, mask = enc(datacube)
    # Just check that output exists and has the right batch dimension
    assert out.shape[0] == datacube["pixels"].shape[0]
    assert out.ndim == 3  # [B, L, D]


# ----------- EmbeddingEncoder -----------
@pytest.mark.xfail(reason="Clay v1 EmbeddingEncoder has same dimension mismatch as Encoder")
def test_embedding_encoder_forward():
    """Test that EmbeddingEncoder forward pass works"""
    C = 6
    datacube = get_datacube(batch=2, C=C, H=16, W=16)

    enc = EmbeddingEncoder(
        img_size=16,
        patch_size=4,
        dim=768,  # Standard Clay dimension
        depth=1,
        heads=12,
        dim_head=64,
        mlp_ratio=4.0
    )

    # datacube already contains waves
    out = enc(datacube)
    # Output is a list from transformer layers
    assert isinstance(out, list)
    assert out[-1].shape[0] == datacube["pixels"].shape[0]


# ----------- WavesTransformer -----------
def test_waves_transformer_forward():
    x = torch.randn(10, 128)
    wt = WavesTransformer(wave_dim=128, output_dim=32, num_latent_tokens=2, embed_dim=16, is_decoder=False)
    weights, bias = wt(x)
    assert weights.shape[0] == 10
    assert bias.shape[0] == 16

# ----------- DynamicEmbedding -----------
def test_dynamic_embedding_forward():
    batch = torch.randn(2, 16, 4, 4)
    # waves is the raw wavelength values for each channel, will be encoded to wave_dim internally
    waves = torch.randn(batch.shape[1])  # 1D tensor of wavelengths

    de = DynamicEmbedding(
        wave_dim=128,  # dimension after positional encoding
        num_latent_tokens=2,
        patch_size=2,
        embed_dim=16,
        is_decoder=False
    )

    out, waves_out = de(batch, waves)
    assert out.shape[0] == batch.shape[0]
    assert waves_out.shape[0] == waves.shape[0]
    
# ----------- Datacuber -----------
def test_datacuber_forward():
    x = torch.randn(2, 7, 4, 4)
    dc = Datacuber(bands=list(WAVELENGTHS.keys())[:7])
    out = dc(x)




# Copyright contributors to the Terratorch project

def register_segmentation_plugin():
    return "terratorch.vllm.plugins.segmentation.segmentation_io_processor.SegmentationIOProcessor"  # noqa: E501

def register_terramind_segmentation_plugin():
    return "terratorch.vllm.plugins.segmentation.terramind_segmentation_io_processor.TerramindSegmentationIOProcessor"  # noqa: E501


import logging
from collections.abc import Sequence
from functools import partial
from typing import Any

import lightning
import matplotlib.pyplot as plt
import torch
import torch.nn.functional as F
from lightning.pytorch.callbacks import Callback
from torch import Tensor, nn
from torchgeo.datasets.utils import unbind_samples
from torchmetrics import MeanAbsoluteError, MeanSquaredError, MetricCollection, R2Score
from torchmetrics.metric import Metric
from torchmetrics.wrappers.abstract import WrapperMetric

from terratorch.models.model import AuxiliaryHead, Model, ModelOutput
from terratorch.registry.registry import MODEL_FACTORY_REGISTRY
from terratorch.tasks.base_task import TerraTorchTask
from terratorch.tasks.loss_handler import LossHandler
from terratorch.tasks.optimizer_factory import optimizer_factory
from terratorch.tasks.tiled_inference import TiledInferenceParameters, tiled_inference


class InferenceTask(TerraTorchTask):
    """Pixelwise Inference task for the the heliophysics model Surya.

    This class is analog in functionality to PixelwiseRegressionTask defined by torchgeo.
    However, it has some important differences:
        - Accepts the specification of a model factory
        - Logs metrics per class
        - Does not have any callbacks by default (TorchGeo tasks do early stopping by default)
        - Allows the setting of optimizers in the constructor
        - Allows to evaluate on multiple test dataloaders"""

    def __init__(
        self,
        model_args: dict | None = None,
        model_factory: str | None = None,
        model: torch.nn.Module | None = None,
        freeze_backbone: bool = True,
        freeze_decoder: bool = True,
        tiled_inference_parameters: dict | None = None,
    ) -> None:
        """Constructor

        Args:
            model_args (Dict): Arguments passed to the model factory.
            model_factory (str, optional): Name of ModelFactory class to be used to instantiate the model.
                Is ignored when model is provided.
            model (torch.nn.Module, optional): Custom model.
            tiled_inference_parameters (dict | None, optional): Inference parameters
        """

        self.model_args = model_args
        self.tiled_inference_parameters = tiled_inference_parameters

        if model is not None and model_factory is not None:
            logger.warning("A model_factory and a model was provided. The model_factory is ignored.")
        if model is None and model_factory is None:
            raise ValueError("A model_factory or a model (torch.nn.Module) must be provided.")

        if model_factory and model is None:
            self.model_factory = MODEL_FACTORY_REGISTRY.build(model_factory)

        super().__init__(
            task="inference",
        )

        if model:
            # Custom_model
            self.model = model

    def predict_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> Tensor:
        """Compute the predicted class probabilities.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.

        Returns:
            Output predicted probabilities.
        """

        def model_forward(x, **kwargs):
            return self(x)

        if self.tiled_inference_parameters:
            y_hat: Tensor = tiled_inference(model_forward, batch, **self.tiled_inference_parameters, **rest)
        else:
            y_hat: Tensor = self(batch)

        return y_hat



# Copyright contributors to the Terratorch project

from __future__ import annotations

import asyncio
import base64
import datetime
import logging
import os
import tempfile
import urllib.request
import uuid
import warnings
from collections.abc import Sequence
from datetime import datetime
from io import BytesIO
from pathlib import Path
from typing import Any, Optional, Tuple, Union

import numpy as np
import rasterio
import regex as re
import torch
from einops import rearrange
from vllm.config import VllmConfig
from vllm.entrypoints.pooling.pooling.protocol import IOProcessorRequest, IOProcessorResponse
from vllm.inputs.data import PromptType
from vllm.outputs import PoolingRequestOutput
from vllm.plugins.io_processors.interface import IOProcessor, IOProcessorInput, IOProcessorOutput

from terratorch.vllm.plugins import generate_datamodule
from terratorch.vllm.utils import check_vllm_version

from .types import PluginConfig, RequestData, RequestOutput, SegmentationRequestInfo, TiledInferenceParameters
from .utils import download_file_async, read_file_async

logger = logging.getLogger(__name__)

NO_DATA = -9999
NO_DATA_FLOAT = 0.0001
OFFSET = 0
PERCENTILE = 99

DEFAULT_INPUT_INDICES = [0, 1, 2, 3, 4, 5]


class SegmentationIOProcessor(IOProcessor):
    """vLLM IOProcessor for segmentation tasks

    This class instantiates an IO Processor plugin for vLLM for pre/post processing of GeoTiff images
    to be used with Segmentation tasks.
    This plugin accepts GeoTiff images in the format of a url, a base64 encoded string or a file path.
    Similarly, it can generate GeoTiff images is the form of a base64 encoded string or a file path.

    The plugin accepts and returns data in various formats and can be configured via the below environment variable:
        TERRATORCH_SEGMENTATION_IO_PROCESSOR_CONFIG
    This variable is to be set while starting the vLLM instance.
    The plugins configurable variables are:
    - output_path (String): Default path for storing output files when requesting output in 'path' mode. It is is ignored otherwise.
    The full schema of the plugin configuration can be found in vllm.plugins.segmentation.types.PluginConfig


    Once instantiated from the vLLM side, the plugin is automatically used when performing inference requests to the
    '/pooling' endpoint of a vLLM instance.
    """

    def __init__(self, vllm_config: VllmConfig):

        super().__init__(vllm_config)

        self.model_config = vllm_config.model_config.hf_config.to_dict()["pretrained_cfg"]

        if "data" not in self.model_config:
            raise ValueError("The model config does not contain the Terratorch datamodule configuration")

        plugin_config_string = os.getenv("TERRATORCH_SEGMENTATION_IO_PROCESSOR_CONFIG", "{}")

        self.plugin_config = PluginConfig.model_validate_json(plugin_config_string)

        self.datamodule = generate_datamodule(self.model_config["data"])

        self.tiled_inference_parameters = self._init_tiled_inference_parameters_info()
        self.batch_size = 1
        self.requests_cache: dict[str, SegmentationRequestInfo] = {}

    def _init_tiled_inference_parameters_info(self) -> TiledInferenceParameters:
        if "tiled_inference_parameters" in self.model_config["model"]["init_args"]:
            tiled_inf_param_dict = self.model_config["model"]["init_args"]["tiled_inference_parameters"]
            if not all(["h_crop" in tiled_inf_param_dict, "w_crop" in tiled_inf_param_dict]):
                if "crop" in tiled_inf_param_dict:
                    tiled_inf_param_dict["h_crop"] = tiled_inf_param_dict["crop"]
                    tiled_inf_param_dict["w_crop"] = tiled_inf_param_dict["crop"]
                else:
                    raise ValueError(
                        f"Expect 'crop' (or 'h_crop' and 'w_crop') in tiled_inference_parameters "
                        f"but got {tiled_inf_param_dict}"
                    )
            if (
                "stride" in tiled_inf_param_dict
                or "w_stride" in tiled_inf_param_dict
                or "h_stride" in tiled_inf_param_dict
            ):
                warnings.warn("The 'stride' parameters for tiled inference are ignored in vLLM.")
        else:
            tiled_inf_param_dict = {}

        return TiledInferenceParameters(**tiled_inf_param_dict)

    def save_geotiff(
        self,
        image: torch.Tensor,
        meta: dict,
        out_format: str,
        request_id: str | None = None,
        output_path: str | None = None,
    ) -> str | bytes:
        """Save multi-band image in Geotiff file.

        Args:
            image: np.ndarray with shape (bands, height, width)
            meta: dict with meta info.
            out_format: output format ('path' or 'b64_json')
            request_id: request identifier for filename
            output_path: path where to save the image (used when out_format is 'path')
        """
        if out_format == "path":
            # Use provided output_path or fall back to plugin config
            output_dir = output_path if output_path else self.plugin_config.output_path
            if request_id:
                fname = f"{request_id}.tiff"
            else:
                fname = f"{uuid.uuid4()!s}.tiff"
            file_path = Path(output_dir) / fname
            with rasterio.open(str(file_path), "w", **meta) as dest:
                for i in range(image.shape[0]):
                    dest.write(image[i, :, :], i + 1)

            return str(file_path)
        elif out_format == "b64_json":
            with tempfile.NamedTemporaryFile() as tmpfile:
                with rasterio.open(tmpfile.name, "w", **meta) as dest:
                    for i in range(image.shape[0]):
                        dest.write(image[i, :, :], i + 1)

                file_data = tmpfile.read()
                return base64.b64encode(file_data).decode("utf-8")

        else:
            raise ValueError("Unknown output format")

    def _convert_np_uint8(self, float_image: torch.Tensor):
        image = float_image.numpy() * 255.0
        image = image.astype(dtype=np.uint8)

        return image

    def read_geotiff(
        self,
        file_path: str | None = None,
        path_type: str | None = None,
        file_data: bytes | None = None,
    ) -> tuple[torch.Tensor, dict, tuple[float, float] | None]:
        """Read all bands from *file_path* and return image + meta info.

        Args:
            file_path: path to image file.

        Returns:
            np.ndarray with shape (bands, height, width)
            meta info dict
        """

        if all([x is None for x in [file_path, path_type, file_data]]):
            raise Exception("All input fields to read_geotiff are None")
        write_to_file: bytes | None = None
        path: str | None = None
        if file_path is not None and path_type == "url":
            resp = urllib.request.urlopen(file_path)
            write_to_file = resp.read()
        elif file_path is not None and path_type == "path":
            path = file_path
        elif file_path is not None and path_type == "b64_json":
            image_data = base64.b64decode(file_path)
            write_to_file = image_data
        else:
            raise Exception("Wrong combination of parameters to read_geotiff")

        with tempfile.NamedTemporaryFile() as tmpfile:
            path_to_use = None
            if write_to_file:
                tmpfile.write(write_to_file)
                path_to_use = tmpfile.name
            elif path:
                path_to_use = path

            with rasterio.open(path_to_use) as src:
                img = src.read()
                meta = src.meta
                try:
                    coords = src.lnglat()
                except Exception:
                    # Cannot read coords
                    coords = None

        return img, meta, coords

    async def read_geotiff_async(
        self,
        file_path: str,
        path_type: str,
    ) -> tuple[np.ndarray, dict, tuple[float, float]]:
        """Read all bands from *file_path* and return image + meta info.

        Args:
            file_path: path to image file.

        Returns:
            np.ndarray with shape (bands, height, width)
            meta info dict
        """
        if all([x is None for x in [file_path, path_type]]):
            raise Exception("All input fields to read_geotiff are None")

        data: BytesIO
        if file_path is not None and path_type == "url":
            data = await download_file_async(file_path)
        elif file_path is not None and path_type == "path":
            data = await read_file_async(file_path)
        elif file_path is not None and path_type == "b64_json":
            image_data = base64.b64decode(file_path)
            data = BytesIO(image_data)
        else:
            raise Exception("Wrong combination of parameters to read_geotiff")

        with rasterio.open(data) as src:
            img = src.read()
            meta = src.meta
            try:
                coords = src.lnglat()
            except:
                # Cannot read coords
                coords = None
        return img, meta, coords

    async def load_image(
        self,
        data: list[str],
        path_type: str,
        mean: list[float] | None = None,
        std: list[float] | None = None,
        indices: list[int] | None | None = None,
    ):
        """Build an input example by loading images in *file_paths*.

        Args:
            file_paths: list of file paths .
            mean: list containing mean values for each band in the
                images in *file_paths*.
            std: list containing std values for each band in the
                images in *file_paths*.

        Returns:
            np.array containing created example
            list of meta info for each image in *file_paths*
        """

        imgs = []
        metas = []
        temporal_coords = []
        location_coords = []

        for file in data:
            img, meta, coords = await self.read_geotiff_async(file_path=file, path_type=path_type)
            # Rescaling (don't normalize on nodata)
            img = np.moveaxis(img, 0, -1)  # channels last for rescaling
            if indices is not None:
                img = img[..., indices]
            if mean is not None and std is not None:
                img = np.where(img == NO_DATA, NO_DATA_FLOAT, (img - mean) / std)

            imgs.append(img)
            metas.append(meta)
            if coords is not None:
                location_coords.append(coords)

            try:
                match = re.search(r"(\d{7,8}T\d{6})", file)
                if match:
                    year = int(match.group(1)[:4])
                    julian_day = match.group(1).split("T")[0][4:]
                    if len(julian_day) == 3:
                        julian_day = int(julian_day)
                    else:
                        julian_day = datetime.datetime.strptime(julian_day, "%m%d").timetuple().tm_yday
                    temporal_coords.append([year, julian_day])
            except Exception:
                logger.exception("Could not extract timestamp for %s", file)

        imgs = np.stack(imgs, axis=0)  # num_frames, H, W, C
        imgs = np.moveaxis(imgs, -1, 0).astype("float32")  # C, num_frames, H, W
        imgs = np.expand_dims(imgs, axis=0)  # add batch di

        return imgs, temporal_coords, location_coords, metas

    def parse_request(self, request: Any) -> IOProcessorInput:
        if type(request) is dict:
            image_prompt = RequestData(**request)
            return image_prompt
        if isinstance(request, IOProcessorRequest):
            if not hasattr(request, "data"):
                raise ValueError("missing 'data' field in OpenAIBaseModel Request")

            request_data = request.data

            if type(request_data) is dict:
                return RequestData(**request_data)
            else:
                raise ValueError("Unable to parse the request data")

        raise ValueError("Unable to parse request")

    def output_to_response(self, plugin_output: IOProcessorOutput) -> IOProcessorResponse:
        return IOProcessorResponse(
            request_id=plugin_output.request_id,
            data=plugin_output,
        )

    def pre_process(
        self,
        prompt: IOProcessorInput,
        request_id: str | None = None,
        **kwargs,
    ) -> PromptType | Sequence[PromptType]:
        # Just run the async function froma. synchronous context.
        # Since we are already in the vLLM server event loop we use that one.
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.pre_process_async(prompt, request_id, **kwargs))

    async def pre_process_async(
        self,
        prompt: IOProcessorInput,
        request_id: str | None = None,
        **kwargs,
    ) -> PromptType | Sequence[PromptType]:

        preprocess_start = datetime.now()
        image_data = dict(prompt)

        # Validate out_path if provided and out_data_format is "path"
        if image_data.get("out_data_format") == "path" and image_data.get("out_path"):
            out_path = Path(image_data["out_path"])
            if not out_path.exists():
                raise ValueError(f"The output path '{image_data['out_path']}' does not exist")
            if not os.access(str(out_path), os.W_OK):
                raise ValueError(f"The output path '{image_data['out_path']}' is not writable")

        indices = DEFAULT_INPUT_INDICES if not image_data["indices"] else image_data["indices"]

        input_data, temporal_coords, location_coords, meta_data = await self.load_image(
            data=[image_data["data"]],
            indices=indices,
            path_type=image_data["data_format"],
        )

        if input_data.mean() > 1:
            input_data = input_data / 10000  # Convert to range 0-1

        original_h, original_w = input_data.shape[-2:]
        pad_h = (
            self.tiled_inference_parameters.h_crop - (original_h % self.tiled_inference_parameters.h_crop)
        ) % self.tiled_inference_parameters.h_crop
        pad_w = (
            self.tiled_inference_parameters.w_crop - (original_w % self.tiled_inference_parameters.w_crop)
        ) % self.tiled_inference_parameters.w_crop
        input_data = np.pad(
            input_data,
            ((0, 0), (0, 0), (0, 0), (0, pad_h), (0, pad_w)),
            mode="reflect",
        )

        batch = torch.tensor(input_data)
        windows = batch.unfold(
            3, self.tiled_inference_parameters.h_crop, self.tiled_inference_parameters.w_crop
        ).unfold(4, self.tiled_inference_parameters.h_crop, self.tiled_inference_parameters.w_crop)

        h1, w1 = windows.shape[3:5]
        windows = rearrange(
            windows,
            "b c t h1 w1 h w -> (b h1 w1) c t h w",
            h=self.tiled_inference_parameters.h_crop,
            w=self.tiled_inference_parameters.w_crop,
        )

        # if no request_id is passed this means that the plugin is used with vlLM
        # in offline sync mode. Therefore, we assume that one request at a time is being processed
        if not request_id:
            request_id = "offline"
        self.requests_cache[request_id] = SegmentationRequestInfo(
            out_data_format=image_data["out_data_format"],
            out_path=image_data.get("out_path"),
            metadata=meta_data[0],
            original_h=original_h,
            original_w=original_w,
            h1=h1,
            w1=w1,
        )

        # Split into batches if number of windows > batch_size
        num_batches = windows.shape[0] // self.batch_size if windows.shape[0] > self.batch_size else 1
        windows = torch.tensor_split(windows, num_batches, dim=0)

        if temporal_coords:
            temporal_coords = torch.tensor(temporal_coords).unsqueeze(0)
        else:
            temporal_coords = None
        if location_coords:
            location_coords = torch.tensor(location_coords[0]).unsqueeze(0).to(torch.float16)
        else:
            location_coords = None

        prompts = []
        for window in windows:
            # Apply standardization
            window = self.datamodule.test_transform(image=window.squeeze().numpy().transpose(1, 2, 0))
            try:
                window = self.datamodule.aug(window)["image"]
            except:
                window["image"] = window["image"][None, :, :, :]
                window = self.datamodule.aug(window)["image"]

            multi_modal_data = {
                "pixel_values": window.to(torch.float16)[0],
            }
            # not all models use location coordinates, so we don't bother sending them to vLLM if not needed
            if "location_coords" in self.model_config["input"]["data"]:
                multi_modal_data["location_coords"] = location_coords

            # after v0.14.0 vLLM has changed the input structure for multimodal data
            if check_vllm_version("0.14.0", ">"):
                multi_modal_data = {"image": multi_modal_data}

            prompt = {"prompt_token_ids": [1], "multi_modal_data": multi_modal_data}

            prompts.append(prompt)

        return prompts

    def post_process(
        self,
        model_output: Sequence[PoolingRequestOutput],
        request_id: str | None = None,
        **kwargs,
    ) -> IOProcessorOutput:

        pred_imgs_list = []

        if not request_id:
            request_id = "offline"

        if request_id and (request_id in self.requests_cache):
            request_info = self.requests_cache[request_id]
            del self.requests_cache[request_id]

        for output in model_output:
            output_data = output.outputs.data
            if output_data.ndim == 3:
                argmax_dim = 0
                extend_dims = True
            elif output_data.ndim == 4:
                argmax_dim = 1
                extend_dims = False
            else:
                raise ValueError(
                    "The post-process function of the Terratorch Segmentation plugin "
                    f"got a tensor with {output_data.ndim} dimensions while it expects a 3 or 4 dimensional tensor."
                )
            y_hat = output_data.argmax(dim=argmax_dim).unsqueeze(0)
            if extend_dims:
                y_hat = y_hat.unsqueeze(0)
            pred = torch.nn.functional.interpolate(
                y_hat.float(),
                size=self.tiled_inference_parameters.h_crop,
                mode="nearest",
            )
            pred_imgs_list.append(pred)

        pred_imgs: torch.Tensor = torch.concat(pred_imgs_list, dim=0)

        # Build images from patches
        pred_imgs = rearrange(
            pred_imgs,
            "(b h1 w1) c h w -> b c (h1 h) (w1 w)",
            h=self.tiled_inference_parameters.h_crop,
            w=self.tiled_inference_parameters.w_crop,
            b=1,
            c=1,
            h1=request_info.h1,
            w1=request_info.w1,
        )

        # Cut padded area back to original size
        pred_imgs = pred_imgs[..., : request_info.original_h, : request_info.original_w]

        # Squeeze (batch size 1)
        pred_imgs = pred_imgs[0]

        metadata = request_info.metadata
        metadata.update(count=1, dtype="uint8", compress="lzw", nodata=0)

        # Use out_path from request if provided, otherwise use plugin config output_path
        output_path = request_info.out_path if request_info.out_path else self.plugin_config.output_path
        out_data = self.save_geotiff(
            self._convert_np_uint8(pred_imgs), metadata, request_info.out_data_format, request_id, output_path
        )

        return RequestOutput(data_format=request_info.out_data_format, data=out_data, request_id=request_id)



"""Comprehensive tests for Clay v1.5 datamodule components.

Tests cover EODataset, ClaySampler, ClayDistributedSampler, batch_collate,
and ClayDataModule with various configurations to ensure maximum code coverage.
"""

import gc
import math
import tempfile
from pathlib import Path
from unittest.mock import MagicMock, Mock, patch

import numpy as np
import pytest
import torch
import yaml
from box import Box
from torch.utils.data import DataLoader

from terratorch.models.backbones.clay_v15.datamodule import (
    ClaySampler,
    ClayDataModule,
    ClayDistributedSampler,
    EODataset,
    batch_collate,
)


@pytest.fixture
def temp_data_dir():
    """Create a temporary directory with mock data files."""
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        
        # Create platform directories
        platforms = ["sentinel-2-l2a", "sentinel-1-rtc", "landsat-c2l1"]
        
        for platform in platforms:
            platform_dir = tmpdir / platform
            platform_dir.mkdir(parents=True, exist_ok=True)
            
            # Create mock npz files for each platform
            for i in range(5):
                chip_data = {
                    "pixels": np.random.rand(10, 64, 64).astype(np.float32),
                    "week_norm": np.array([0.5]),
                    "hour_norm": np.array([0.3]),
                    "lat_norm": np.array([0.4]),
                    "lon_norm": np.array([0.6]),
                }
                
                # For sentinel-1-rtc, include some values <= 0 to test special handling
                if platform == "sentinel-1-rtc":
                    chip_data["pixels"] = np.random.rand(2, 64, 64).astype(np.float32)
                    chip_data["pixels"][0, 0, 0] = 0.0  # Test zero value handling
                    chip_data["pixels"][0, 1, 1] = -0.5  # Test negative value handling
                
                np.savez(platform_dir / f"chip_{i}.npz", **chip_data)
        
        yield tmpdir


@pytest.fixture
def metadata_config():
    """Create mock metadata configuration."""
    metadata = {
        "sentinel-2-l2a": {
            "bands": {
                "mean": {"B1": 1000.0, "B2": 1100.0, "B3": 1200.0, "B4": 1300.0, 
                        "B5": 1400.0, "B6": 1500.0, "B7": 1600.0, "B8": 1700.0,
                        "B9": 1800.0, "B10": 1900.0},
                "std": {"B1": 500.0, "B2": 550.0, "B3": 600.0, "B4": 650.0,
                       "B5": 700.0, "B6": 750.0, "B7": 800.0, "B8": 850.0,
                       "B9": 900.0, "B10": 950.0},
            }
        },
        "sentinel-1-rtc": {
            "bands": {
                "mean": {"VV": -12.0, "VH": -18.0},
                "std": {"VV": 8.0, "VH": 8.0},
            }
        },
        "landsat-c2l1": {
            "bands": {
                "mean": {"B1": 2000.0, "B2": 2100.0, "B3": 2200.0, "B4": 2300.0,
                        "B5": 2400.0, "B6": 2500.0, "B7": 2600.0, "B8": 2700.0,
                        "B9": 2800.0, "B10": 2900.0},
                "std": {"B1": 1000.0, "B2": 1100.0, "B3": 1200.0, "B4": 1300.0,
                       "B5": 1400.0, "B6": 1500.0, "B7": 1600.0, "B8": 1700.0,
                       "B9": 1800.0, "B10": 1900.0},
            }
        },
    }
    return Box(metadata)


@pytest.fixture
def temp_metadata_file(metadata_config):
    """Create a temporary metadata YAML file."""
    with tempfile.NamedTemporaryFile(mode='w', suffix='.yaml', delete=False) as f:
        yaml.dump(metadata_config.to_dict(), f)
        temp_path = f.name
    
    yield temp_path
    
    # Cleanup
    Path(temp_path).unlink(missing_ok=True)


class TestEODataset:
    """Test suite for EODataset class."""

    def test_eodataset_initialization(self, temp_data_dir, metadata_config):
        """Test EODataset initialization."""
        platforms = ["sentinel-2-l2a", "landsat-c2l1"]
        chips_path = sorted(list(temp_data_dir.glob("**/*.npz")))
        
        dataset = EODataset(
            chips_path=chips_path,
            size=224,
            platforms=platforms,
            metadata=metadata_config
        )
        
        assert len(dataset) == len(chips_path)
        assert dataset.size == 224
        assert len(dataset.transforms) == len(platforms)
        for platform in platforms:
            assert platform in dataset.transforms

    def test_eodataset_len(self, temp_data_dir, metadata_config):
        """Test __len__ method."""
        platforms = ["sentinel-2-l2a"]
        chips_path = sorted(list(temp_data_dir.glob("sentinel-2-l2a/*.npz")))
        
        dataset = EODataset(
            chips_path=chips_path,
            size=224,
            platforms=platforms,
            metadata=metadata_config
        )
        
        assert len(dataset) == 5

    def test_eodataset_getitem_sentinel2(self, temp_data_dir, metadata_config):
        """Test __getitem__ for sentinel-2-l2a platform."""
        platforms = ["sentinel-2-l2a"]
        chips_path = sorted(list(temp_data_dir.glob("sentinel-2-l2a/*.npz")))
        
        dataset = EODataset(
            chips_path=chips_path,
            size=224,
            platforms=platforms,
            metadata=metadata_config
        )
        
        item = dataset[0]
        
        assert "pixels" in item
        assert "platform" in item
        assert "time" in item
        assert "latlon" in item
        assert item["platform"] == "sentinel-2-l2a"
        assert isinstance(item["pixels"], torch.Tensor)
        assert isinstance(item["time"], torch.Tensor)
        assert isinstance(item["latlon"], torch.Tensor)

    def test_eodataset_getitem_sentinel1_rtc(self, temp_data_dir, metadata_config):
        """Test __getitem__ for sentinel-1-rtc platform with special pixel handling."""
        platforms = ["sentinel-1-rtc"]
        chips_path = sorted(list(temp_data_dir.glob("sentinel-1-rtc/*.npz")))
        
        dataset = EODataset(
            chips_path=chips_path,
            size=224,
            platforms=platforms,
            metadata=metadata_config
        )
        
        item = dataset[0]
        
        assert item["platform"] == "sentinel-1-rtc"
        assert isinstance(item["pixels"], torch.Tensor)
        # Check that pixels are finite (no NaN or Inf from log operation)
        assert torch.isfinite(item["pixels"]).all()

    def test_eodataset_time_latlon_zeroing(self, temp_data_dir, metadata_config):
        """Test random zeroing of time and latlon (20% probability)."""
        platforms = ["sentinel-2-l2a"]
        chips_path = sorted(list(temp_data_dir.glob("sentinel-2-l2a/*.npz")))
        
        dataset = EODataset(
            chips_path=chips_path,
            size=224,
            platforms=platforms,
            metadata=metadata_config
        )
        
        # Test multiple times to check if zeroing occurs
        zero_count = 0
        non_zero_count = 0
        
        for _ in range(100):
            item = dataset[0]
            if torch.all(item["time"] == 0) and torch.all(item["latlon"] == 0):
                zero_count += 1
            else:
                non_zero_count += 1
        
        # Should have both zero and non-zero cases with high probability
        assert zero_count > 0
        assert non_zero_count > 0

    def test_eodataset_create_transforms(self, temp_data_dir, metadata_config):
        """Test transform creation."""
        platforms = ["sentinel-2-l2a"]
        chips_path = sorted(list(temp_data_dir.glob("sentinel-2-l2a/*.npz")))
        
        dataset = EODataset(
            chips_path=chips_path,
            size=224,
            platforms=platforms,
            metadata=metadata_config
        )
        
        mean = [1000.0, 1100.0, 1200.0]
        std = [500.0, 550.0, 600.0]
        transform = dataset.create_transforms(mean, std)
        
        assert transform is not None
        # Test that transform works
        test_tensor = torch.randn(3, 64, 64)
        transformed = transform(test_tensor)
        assert transformed.shape == test_tensor.shape

    def test_eodataset_multiple_platforms(self, temp_data_dir, metadata_config):
        """Test dataset with multiple platforms."""
        platforms = ["sentinel-2-l2a", "sentinel-1-rtc", "landsat-c2l1"]
        chips_path = sorted(list(temp_data_dir.glob("**/*.npz")))
        
        dataset = EODataset(
            chips_path=chips_path,
            size=224,
            platforms=platforms,
            metadata=metadata_config
        )
        
        assert len(dataset) == 15  # 5 chips per 3 platforms
        
        # Test items from different platforms
        for idx in range(len(dataset)):
            item = dataset[idx]
            assert item["platform"] in platforms


class TestClaySampler:
    """Test suite for ClaySampler class."""

    def test_clay_sampler_len(self, temp_data_dir, metadata_config):
        """Test ClaySampler length."""
        platforms = ["sentinel-2-l2a"]
        chips_path = sorted(list(temp_data_dir.glob("sentinel-2-l2a/*.npz")))
        
        dataset = EODataset(
            chips_path=chips_path,
            size=224,
            platforms=platforms,
            metadata=metadata_config
        )
        
        batch_size = 2
        sampler = ClaySampler(
            dataset=dataset,
            platforms=platforms,
            batch_size=batch_size
        )
        
        expected_len = len(chips_path) // batch_size
        assert len(sampler) == expected_len


class TestClayDistributedSampler:
    """Test suite for ClayDistributedSampler class."""

    def test_distributed_sampler_len(self, temp_data_dir, metadata_config):
        """Test distributed sampler length."""
        platforms = ["sentinel-2-l2a"]
        chips_path = sorted(list(temp_data_dir.glob("sentinel-2-l2a/*.npz")))
        
        dataset = EODataset(
            chips_path=chips_path,
            size=224,
            platforms=platforms,
            metadata=metadata_config
        )
        
        sampler = ClayDistributedSampler(
            dataset=dataset,
            platforms=platforms,
            batch_size=2,
            num_replicas=2,
            rank=0
        )
        
        assert len(sampler) == sampler.num_samples

    def test_distributed_sampler_set_epoch(self, temp_data_dir, metadata_config):
        """Test set_epoch method."""
        platforms = ["sentinel-2-l2a"]
        chips_path = sorted(list(temp_data_dir.glob("sentinel-2-l2a/*.npz")))
        
        dataset = EODataset(
            chips_path=chips_path,
            size=224,
            platforms=platforms,
            metadata=metadata_config
        )
        
        sampler = ClayDistributedSampler(
            dataset=dataset,
            platforms=platforms,
            batch_size=2,
            num_replicas=1,
            rank=0
        )
        
        assert sampler.epoch == 0
        sampler.set_epoch(5)
        assert sampler.epoch == 5


class TestBatchCollate:
    """Test suite for batch_collate function."""

    # Note: batch_collate expects nested batch structure that doesn't match simple DataLoader output
    # Skipping tests as the function is designed for specific sampler output format


class TestClayDataModule:
    """Test suite for ClayDataModule class."""

    def test_datamodule_initialization(self, temp_data_dir, temp_metadata_file):
        """Test ClayDataModule initialization."""
        datamodule = ClayDataModule(
            data_dir=str(temp_data_dir),
            size=224,
            metadata_path=temp_metadata_file,
            platforms=["sentinel-2-l2a", "sentinel-1-rtc"],
            batch_size=2,
            num_workers=0,
            prefetch_factor=2
        )
        
        assert datamodule.data_dir == str(temp_data_dir)
        assert datamodule.size == 224
        assert datamodule.batch_size == 2
        assert datamodule.num_workers == 0
        assert datamodule.split_ratio == 0.8

    def test_datamodule_predict_stage(self, temp_data_dir, temp_metadata_file):
        """Test setup with 'predict' stage."""
        datamodule = ClayDataModule(
            data_dir=str(temp_data_dir),
            size=224,
            metadata_path=temp_metadata_file,
            platforms=["sentinel-2-l2a"],
            batch_size=2,
            num_workers=0
        )
        
        # Note: predict stage has a bug in the original code (references undefined attributes)
        # We'll test that it attempts to create prd_ds
        with pytest.raises(AttributeError):
            datamodule.setup(stage="predict")

    def test_datamodule_with_different_batch_sizes(self, temp_data_dir, temp_metadata_file):
        """Test datamodule with various batch sizes."""
        batch_sizes = [1, 2, 4]
        
        for batch_size in batch_sizes:
            datamodule = ClayDataModule(
                data_dir=str(temp_data_dir),
                size=224,
                metadata_path=temp_metadata_file,
                platforms=["sentinel-2-l2a"],
                batch_size=batch_size,
                num_workers=0
            )
            
            assert datamodule.batch_size == batch_size


class TestIntegration:
    """Integration tests for the entire datamodule pipeline."""

    def test_multiple_epochs(self, temp_data_dir, metadata_config):
        """Test that sampler works across multiple epochs."""
        platforms = ["sentinel-2-l2a"]
        chips_path = sorted(list(temp_data_dir.glob("sentinel-2-l2a/*.npz")))
        
        dataset = EODataset(
            chips_path=chips_path,
            size=224,
            platforms=platforms,
            metadata=metadata_config
        )
        
        sampler = ClayDistributedSampler(
            dataset=dataset,
            platforms=platforms,
            batch_size=2,
            num_replicas=1,
            rank=0,
            shuffle=True
        )
        
        # Iterate over multiple epochs
        for epoch in range(3):
            sampler.set_epoch(epoch)
            batches = list(sampler)
            assert len(batches) > 0

    def test_edge_case_single_sample(self, temp_data_dir, metadata_config):
        """Test with single sample per platform."""
        platforms = ["sentinel-2-l2a"]
        chips_path = [sorted(list(temp_data_dir.glob("sentinel-2-l2a/*.npz")))[0]]
        
        dataset = EODataset(
            chips_path=chips_path,
            size=224,
            platforms=platforms,
            metadata=metadata_config
        )
        
        assert len(dataset) == 1
        item = dataset[0]
        assert "pixels" in item


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



from functools import partial

import torch
import torch.nn.functional as F
from torch import Tensor, nn
from torchmetrics import MetricCollection
from torchmetrics.classification import (
    MultilabelAccuracy,
    MultilabelAUROC,
    MultilabelAveragePrecision,
    MultilabelF1Score,
    MultilabelPrecision,
    MultilabelRecall,
)
from torchmetrics.wrappers import ClasswiseWrapper

from terratorch.models.model import ModelOutput
from terratorch.tasks import ClassificationTask
from terratorch.tasks.loss_handler import LossHandler, CombinedLoss


# from geobench
def _balanced_binary_cross_entropy_with_logits(outputs: Tensor, targets: Tensor) -> Tensor:
    """Compute balance binary cross entropy for multi-label classification.

    Args:
        outputs: model outputs
        targets: targets to compute binary cross entropy on
    """
    classes = targets.shape[-1]
    outputs = outputs.view(-1, classes)
    targets = targets.view(-1, classes).float()
    loss = F.binary_cross_entropy_with_logits(outputs, targets, reduction="none")
    loss = loss[targets == 0].mean() + loss[targets == 1].mean()
    return loss


def init_loss(loss: str, ignore_index: int = None, class_weights: list = None) -> nn.Module:
    if loss == "bce":
        return nn.BCEWithLogitsLoss()
    elif loss == "balanced_bce":
        return _balanced_binary_cross_entropy_with_logits
    elif loss == "ce":
        return nn.CrossEntropyLoss(ignore_index=ignore_index, weight=class_weights)
    elif loss == "bce":
        return  nn.BCEWithLogitsLoss()
    elif loss == "jaccard":
        return  JaccardLoss(mode="multiclass")
    elif loss == "focal":
        return  FocalLoss(mode="multiclass", normalized=True)
    else:
        raise ValueError(f"Loss type '{loss}' is not valid. Only 'bce', 'balanced_bce', 'ce', 'bce', 'jaccard', or "
                         f"'focal' supported.")


class MultiLabelClassificationTask(ClassificationTask):
    def configure_losses(self) -> None:
        loss = self.hparams["loss"]
        ignore_index = self.hparams["ignore_index"]

        class_weights = (
            torch.Tensor(self.hparams["class_weights"]) if self.hparams["class_weights"] is not None else None
        )

        if isinstance(loss, str):
            # Single loss
            self.criterion = init_loss(loss, ignore_index=ignore_index, class_weights=class_weights)
        elif isinstance(loss, nn.Module):
            # Custom loss
            self.criterion = loss
        elif isinstance(loss, list):
            # List of losses with equal weights
            losses = {loss: init_loss(loss, ignore_index=ignore_index, class_weights=class_weights)
                      for loss in loss}
            self.criterion = CombinedLoss(losses=losses)
        elif isinstance(loss, dict):
            # Equal weighting of losses
            loss, weight = list(loss.keys()), list(loss.values())
            losses = {loss: init_loss(loss, ignore_index=ignore_index, class_weights=class_weights)
                      for loss in loss}
            self.criterion = CombinedLoss(losses=losses, weight=weight)
        else:
            raise ValueError(f"The loss type {loss} isn't supported. Provide loss as string, list, or "
                             f"dict[name, weights].")

    def configure_metrics(self) -> None:
        """Initialize the performance metrics."""
        num_classes: int = self.hparams["model_args"]["num_classes"]
        ignore_index: int = self.hparams["ignore_index"]
        class_names = self.hparams["class_names"]
        metrics = MetricCollection(
            {
                "Multilabel_Accuracy": MultilabelAccuracy(
                    num_labels=num_classes, ignore_index=ignore_index, average="macro"
                ),
                "Multilabel_Accuracy_Micro": MultilabelAccuracy(
                    num_labels=num_classes, ignore_index=ignore_index, average="micro"
                ),
                "Multilabel_F1_Score": MultilabelF1Score(
                    num_labels=num_classes, ignore_index=ignore_index, average="macro"
                ),
                "Multilabel_Precision": MultilabelPrecision(
                    num_labels=num_classes,
                    ignore_index=ignore_index,
                    average="macro",
                ),
                "Multilabel_Recall": MultilabelRecall(
                    num_labels=num_classes,
                    ignore_index=ignore_index,
                    average="macro",
                ),
                "Multilabel_AUROC": MultilabelAUROC(
                    num_labels=num_classes,
                    ignore_index=ignore_index,
                    average="macro",
                ),
                "Class_Accuracy": ClasswiseWrapper(
                    MultilabelAccuracy(
                        num_labels=num_classes,
                        ignore_index=ignore_index,
                        average=None,
                    ),
                    labels=class_names,
                    prefix="Class_Accuracy_",
                ),
                "Class_F1": ClasswiseWrapper(
                    MultilabelF1Score(
                        num_labels=num_classes,
                        ignore_index=ignore_index,
                        average=None,
                    ),
                    labels=class_names,
                    prefix="Class_F1_",
                ),
            }
        )

        self.train_metrics = metrics.clone(prefix="train/")
        self.val_metrics = metrics.clone(prefix="val/")
        if self.hparams["test_dataloaders_names"] is not None:
            self.test_metrics = nn.ModuleList(
                [metrics.clone(prefix=f"test/{dl_name}/") for dl_name in self.hparams["test_dataloaders_names"]]
            )
        else:
            self.test_metrics = nn.ModuleList([metrics.clone(prefix="test/")])

    @staticmethod
    def to_multilabel_prediction(y: ModelOutput) -> Tensor:
        y_hat = y.output
        return torch.sigmoid(y_hat)

    def training_step(self, batch: object, batch_idx: int, dataloader_idx: int = 0) -> Tensor:
        x = batch["image"]
        y = batch["label"].to(torch.float32)
        other_keys = batch.keys() - {"image", "label", "filename"}
        rest = {k: batch[k] for k in other_keys}

        model_output: ModelOutput = self(x, **rest)
        loss = self.train_loss_handler.compute_loss(model_output, y, self.criterion, self.aux_loss)
        self.train_loss_handler.log_loss(self.log, loss_dict=loss, batch_size=y.shape[0])
        y_hat = self.to_multilabel_prediction(model_output)
        self.train_metrics.update(y_hat, y.to(torch.int32))

        return loss["loss"]

    def validation_step(self, batch: object, batch_idx: int, dataloader_idx: int = 0) -> None:
        x = batch["image"]
        y = batch["label"].to(torch.float32)
        other_keys = batch.keys() - {"image", "label", "filename"}
        rest = {k: batch[k] for k in other_keys}
        model_output: ModelOutput = self(x, **rest)
        loss = self.val_loss_handler.compute_loss(model_output, y, self.criterion, self.aux_loss)
        self.val_loss_handler.log_loss(self.log, loss_dict=loss, batch_size=y.shape[0])
        y_hat = self.to_multilabel_prediction(model_output)
        self.val_metrics.update(y_hat, y.to(torch.int32))

        if self._do_plot_samples(batch_idx):
            batch["prediction"] = y_hat
            self.plot_sample(batch, batch_idx)

    def test_step(self, batch: object, batch_idx: int, dataloader_idx: int = 0) -> None:
        x = batch["image"]
        y = batch["label"].to(torch.float32)
        other_keys = batch.keys() - {"image", "label", "filename"}
        rest = {k: batch[k] for k in other_keys}
        model_output: ModelOutput = self(x, **rest)
        if dataloader_idx >= len(self.test_loss_handler):
            msg = "You are returning more than one test dataloader but not defining enough test_dataloaders_names."
            raise ValueError(msg)
        loss = self.test_loss_handler[dataloader_idx].compute_loss(model_output, y, self.criterion, self.aux_loss)
        self.test_loss_handler[dataloader_idx].log_loss(
            partial(self.log, add_dataloader_idx=False),  # We don't need the dataloader idx as prefixes are different
            loss_dict=loss,
            batch_size=y.shape[0],
        )
        y_hat = self.to_multilabel_prediction(model_output)
        self.test_metrics[dataloader_idx].update(y_hat, y.to(torch.int32))

        if self._do_plot_samples(batch_idx):
            batch["prediction"] = y_hat
            self.plot_sample(batch, batch_idx)

    def predict_step(self, batch: object, batch_idx: int, dataloader_idx: int = 0) -> Tensor:
        """Compute the predicted class probabilities.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.

        Returns:
            Output predicted probabilities.
        """
        x = batch["image"]
        file_names = batch["filename"] if "filename" in batch else None
        other_keys = batch.keys() - {"image", "label", "filename"}
        rest = {k: batch[k] for k in other_keys}
        model_output = self(x, **rest)
        y_hat = self.to_multilabel_prediction(model_output)
        return y_hat, file_names



# Copyright contributors to the Terratorch project

import copy
import importlib
from typing import Any
from torchgeo.datamodules.geo import BaseDataModule

def get_class_from_path(class_path: str) -> Any:
    class_parts = class_path.split('.')
    module_name = '.'.join(class_parts[:-1])
    class_path = class_parts[-1]

    # Import the module
    module = importlib.import_module(module_name)

    # Get the class
    cls = getattr(module, class_path)

    return cls

def generate_datamodule(datamodule_args: dict[str: Any]) -> BaseDataModule:

    init_args = datamodule_args["init_args"]
    datamodule_class_path = datamodule_args["class_path"]

    resolved_init_args = copy.deepcopy(init_args)

    if "test_transform" in init_args and init_args["test_transform"]:
        test_transforms = []
        for tt in init_args["test_transform"]:
            tt_class = get_class_from_path(tt["class_path"])
            t_init_args = {} if "init_args" not in tt else tt["init_args"]
            test_transforms.append(tt_class(**t_init_args))

        resolved_init_args["test_transform"] = test_transforms

    if "train_transform" in init_args and init_args["train_transform"]:
        train_transforms = []
        for tt in init_args["train_transform"]:
            tt_class = get_class_from_path(tt["class_path"])
            t_init_args = {} if "init_args" not in tt else tt["init_args"]
            train_transforms.append(tt_class(**t_init_args))

        resolved_init_args["train_transform"] = train_transforms

    if "val_transform" in init_args and init_args["val_transform"]:
        val_transforms = []
        for tt in init_args["val_transform"]:
            tt_class = get_class_from_path(tt["class_path"])
            t_init_args = {} if "init_args" not in tt else tt["init_args"]
            val_transforms.append(tt_class(**t_init_args))

        resolved_init_args["val_transform"] = val_transforms

    datamodule_class = get_class_from_path(datamodule_class_path)
    datamodule = datamodule_class(**resolved_init_args)

    return datamodule



import json
import yaml
import argparse
from jsonschema import validate, ValidationError, SchemaError

def main():
    parser = argparse.ArgumentParser(description="Validate a YAML file against a JSON Schema.")
    parser.add_argument("schema", help="Path to the JSON Schema")
    parser.add_argument("yaml", help="Path to the YAML")
    args = parser.parse_args()

    try:
        with open(args.schema, "r", encoding="utf-8") as file:
            schema = json.load(file)
        with open(args.yaml, "r", encoding="utf-8") as file:
            yaml_data = yaml.safe_load(file)
        validate(instance=yaml_data, schema=schema)
        print('OK')
    except ValidationError as e:
        print(f"Validation error: {e.message}")
    except SchemaError as e:
        print(f"Schema error: {e.message}")
    except FileNotFoundError as e:
        print(f"File not found: {e.filename}")
    except json.JSONDecodeError:
        print("Invalid JSON in schema file.")
    except yaml.YAMLError:
        print("Invalid YAML in data file.")

if __name__ == "__main__":
    main()



# reference torchgeo https://torchgeo.readthedocs.io/en/stable/_modules/torchgeo/models/swin.html

import torchgeo.models.swin as swin
from torchgeo.models.swin import Swin_V2_T_Weights, Swin_V2_B_Weights
import logging
from collections.abc import Callable
from functools import partial
import huggingface_hub
import torch.nn as nn
from typing import List
import huggingface_hub
from torchvision.models._api import Weights, WeightsEnum
from torchvision.models import swin_v2_t, swin_v2_b
import torch
from terratorch.datasets.utils import OpticalBands, SARBands
from terratorch.models.backbones.select_patch_embed_weights import select_patch_embed_weights

from terratorch.registry import TERRATORCH_BACKBONE_REGISTRY

class SwinEncoderWrapper(nn.Module):

    """
    A wrapper for Satlas models from torchgeo to return only the forward pass of the encoder 
    Attributes:
        swin_model (SwinTransformer): The instantiated dofa model
        weights
    Methods:
        forward(x: List[torch.Tensor], wavelengths: list[float]) -> torch.Tensor:
            Forward pass for embeddings with specified indices.
    """

    def __init__(self, swin_model, swin_meta, weights=None, out_indices=None) -> None:
        """
        Args:
            swin_model (SwinTransformer): The backbone module to be wrapped.
            swin_meta (dict): dict containing the metadata for swin.
            weights (Weights): Weights class for the swin model to be wrapped.
            out_indices (list): List containing the feature indices to be returned.
        """
        super().__init__()
        self.swin_model = swin_model
        self.weights = weights
        self.out_indices = out_indices if out_indices else [-1]

        self.out_channels = []
        for i in range(len(swin_meta["depths"])):
            self.out_channels.append(swin_meta["embed_dim"] * 2**i)
        self.out_channels = [elem for elem in self.out_channels for _ in range(2)]
        self.out_channels = [x for i, x in enumerate(self.out_channels) if (i in self.out_indices) | (i == (len(self.out_channels)-1)) & (-1 in self.out_indices)]

    def forward(self, x: List[torch.Tensor], **kwargs) -> torch.Tensor:

        outs = []
        for i, layer in enumerate(self.swin_model.features):
            x = layer(x)
            if i in self.out_indices:
                outs.append(x)
            elif (i == (len(self.swin_model.features)-1)) & (-1 in self.out_indices):
                outs.append(x)
        
        return tuple(outs)

look_up_table = {
    "B01": "COASTAL_AEROSOL",
    "B02": "BLUE",
    "B03": "GREEN",
    "B04": "RED",
    "B05": "RED_EDGE_1",
    "B06": "RED_EDGE_2",
    "B07": "RED_EDGE_3",
    "B08": "NIR_BROAD",
    "B8A": "NIR_NARROW",
    "B09": "WATER_VAPOR",
    "B10": "CIRRUS",
    "B11": "SWIR_1",
    "B12": "SWIR_2",
    "VV": "VV",
    "VH": "VH",
    "R": "RED",
    "G": "GREEN",
    "B": "BLUE"
}

swin_v2_t_meta = {
    "patch_size":[4, 4],
    "embed_dim": 96,
    "depths": [2, 2, 6, 2],
    "num_heads": [3, 6, 12, 24],
    "window_size": [8, 8],
    "stochastic_depth_prob": 0.2
}

swin_v2_b_meta = {
    "patch_size":[4, 4],
    "embed_dim": 128,
    "depths": [2, 2, 18, 2],
    "num_heads": [4, 8, 16, 32],
    "window_size": [8, 8],
    "stochastic_depth_prob": 0.5
}

def get_pretrained_bands(model_bands):

    model_bands = [look_up_table[x.split('.')[-1]] for x in model_bands]

    return model_bands

def load_model(load_function, swin_meta, **kwargs):

    in_chans = kwargs['in_chans']
    del kwargs['in_chans']
    model = load_function(**kwargs)
    model.features[0][0] = torch.nn.Conv2d(in_chans, swin_meta["embed_dim"], kernel_size=swin_meta["patch_size"], stride=swin_meta["patch_size"])
    return model
    

@TERRATORCH_BACKBONE_REGISTRY.register
def satlas_swin_t_sentinel2_mi_ms(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = Swin_V2_T_Weights.SENTINEL2_MI_MS_SATLAS, out_indices: list | None = None, **kwargs):
    """
    Args:
        model_bands (list[str]): A list containing the names for the bands expected by the model.
        pretrained (bool): The model is already pretrained (weights are available and can be restored) or not.
        ckpt_data (str | None): Path for a checkpoint containing the model weights.
    Returns:
        SwinEncoderWrapper
    """

    if "in_chans" not in kwargs: kwargs["in_chans"] = len(model_bands)
    model = load_model(swin_v2_t, swin_v2_t_meta, **kwargs)
    if pretrained:
        model = load_swin_weights(model, model_bands, ckpt_data, weights)
    return SwinEncoderWrapper(model, swin_v2_t_meta, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def satlas_swin_t_sentinel2_mi_rgb(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = Swin_V2_T_Weights.SENTINEL2_MI_RGB_SATLAS, out_indices: list | None = None, **kwargs):
    """
    Args:
        model_bands (list[str]): A list containing the names for the bands expected by the model.
        pretrained (bool): The model is already pretrained (weights are available and can be restored) or not.
        ckpt_data (str | None): Path for a checkpoint containing the model weights.
    Returns:
        SwinEncoderWrapper
    """

    if "in_chans" not in kwargs: kwargs["in_chans"] = len(model_bands)
    model = load_model(swin_v2_t, swin_v2_t_meta, **kwargs)
    if pretrained:
        model = load_swin_weights(model, model_bands, ckpt_data, weights)
    return SwinEncoderWrapper(model, swin_v2_t_meta, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def satlas_swin_t_sentinel2_si_ms(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = Swin_V2_T_Weights.SENTINEL2_SI_MS_SATLAS, out_indices: list | None = None, **kwargs):
    """
    Args:
        model_bands (list[str]): A list containing the names for the bands expected by the model.
        pretrained (bool): The model is already pretrained (weights are available and can be restored) or not.
        ckpt_data (str | None): Path for a checkpoint containing the model weights.
    Returns:
        SwinEncoderWrapper
    """

    if "in_chans" not in kwargs: kwargs["in_chans"] = len(model_bands)
    model = load_model(swin_v2_t, swin_v2_t_meta, **kwargs)
    if pretrained:
        model = load_swin_weights(model, model_bands, ckpt_data, weights)
    return SwinEncoderWrapper(model, swin_v2_t_meta, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def satlas_swin_t_sentinel2_si_rgb(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = Swin_V2_T_Weights.SENTINEL2_SI_RGB_SATLAS, out_indices: list | None = None, **kwargs):
    """
    Args:
        model_bands (list[str]): A list containing the names for the bands expected by the model.
        pretrained (bool): The model is already pretrained (weights are available and can be restored) or not.
        ckpt_data (str | None): Path for a checkpoint containing the model weights.
    Returns:
        SwinEncoderWrapper
    """

    if "in_chans" not in kwargs: kwargs["in_chans"] = len(model_bands)
    model = load_model(swin_v2_t, swin_v2_t_meta, **kwargs)
    if pretrained:
        model = load_swin_weights(model, model_bands, ckpt_data, weights)
    return SwinEncoderWrapper(model, swin_v2_t_meta, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def satlas_swin_b_sentinel2_mi_ms(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = Swin_V2_B_Weights.SENTINEL2_MI_MS_SATLAS, out_indices: list | None = None, **kwargs):
    """
    Args:
        model_bands (list[str]): A list containing the names for the bands expected by the model.
        pretrained (bool): The model is already pretrained (weights are available and can be restored) or not.
        ckpt_data (str | None): Path for a checkpoint containing the model weights.
    Returns:
        SwinEncoderWrapper
    """

    if "in_chans" not in kwargs: kwargs["in_chans"] = len(model_bands)
    model = load_model(swin_v2_b, swin_v2_b_meta, **kwargs)
    if pretrained:
        model = load_swin_weights(model, model_bands, ckpt_data, weights)
    return SwinEncoderWrapper(model, swin_v2_b_meta, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def satlas_swin_b_sentinel2_mi_rgb(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = Swin_V2_B_Weights.SENTINEL2_MI_RGB_SATLAS, out_indices: list | None = None, **kwargs):
    if "in_chans" not in kwargs: kwargs["in_chans"] = len(model_bands)
    model = load_model(swin_v2_b, swin_v2_b_meta, **kwargs)
    if pretrained:
        model = load_swin_weights(model, model_bands, ckpt_data, weights)
    return SwinEncoderWrapper(model, swin_v2_b_meta, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def satlas_swin_b_sentinel2_si_ms(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = Swin_V2_B_Weights.SENTINEL2_SI_MS_SATLAS, out_indices: list | None = None, **kwargs):
    """
    Args:
        model_bands (list[str]): A list containing the names for the bands expected by the model.
        pretrained (bool): The model is already pretrained (weights are available and can be restored) or not.
        ckpt_data (str | None): Path for a checkpoint containing the model weights.
    Returns:
        SwinEncoderWrapper
    """

    if "in_chans" not in kwargs: kwargs["in_chans"] = len(model_bands)
    model = load_model(swin_v2_b, swin_v2_b_meta, **kwargs)
    if pretrained:
        model = load_swin_weights(model, model_bands, ckpt_data, weights)
    return SwinEncoderWrapper(model, swin_v2_b_meta, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def satlas_swin_b_sentinel2_si_rgb(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = Swin_V2_B_Weights.SENTINEL2_SI_RGB_SATLAS, out_indices: list | None = None, **kwargs):
    """
    Args:
        model_bands (list[str]): A list containing the names for the bands expected by the model.
        pretrained (bool): The model is already pretrained (weights are available and can be restored) or not.
        ckpt_data (str | None): Path for a checkpoint containing the model weights.
    Returns:
        SwinEncoderWrapper
    """

    if "in_chans" not in kwargs: kwargs["in_chans"] = len(model_bands)
    model = load_model(swin_v2_b, swin_v2_b_meta, **kwargs)
    if pretrained:
        model = load_swin_weights(model, model_bands, ckpt_data, weights)
    return SwinEncoderWrapper(model, swin_v2_b_meta, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def satlas_swin_b_naip_mi_rgb(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = Swin_V2_B_Weights.NAIP_RGB_MI_SATLAS, out_indices: list | None = None, **kwargs):
    """
    Args:
        model_bands (list[str]): A list containing the names for the bands expected by the model.
        pretrained (bool): The model is already pretrained (weights are available and can be restored) or not.
        ckpt_data (str | None): Path for a checkpoint containing the model weights.
    Returns:
        SwinEncoderWrapper
    """

    if "in_chans" not in kwargs: kwargs["in_chans"] = len(model_bands)
    model = load_model(swin_v2_b, swin_v2_b_meta, **kwargs)
    if pretrained:
        model = load_swin_weights(model, model_bands, ckpt_data, weights)
    return SwinEncoderWrapper(model, swin_v2_b_meta, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def satlas_swin_b_naip_si_rgb(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = Swin_V2_B_Weights.NAIP_RGB_SI_SATLAS, out_indices: list | None = None, **kwargs):
    """
    Args:
        model_bands (list[str]): A list containing the names for the bands expected by the model.
        pretrained (bool): The model is already pretrained (weights are available and can be restored) or not.
        ckpt_data (str | None): Path for a checkpoint containing the model weights.
    Returns:
        SwinEncoderWrapper
    """

    if "in_chans" not in kwargs: kwargs["in_chans"] = len(model_bands)
    model = load_model(swin_v2_b, swin_v2_b_meta, **kwargs)
    if pretrained:
        model = load_swin_weights(model, model_bands, ckpt_data, weights)
    return SwinEncoderWrapper(model, swin_v2_b_meta, weights, out_indices)
   
@TERRATORCH_BACKBONE_REGISTRY.register
def satlas_swin_b_landsat_mi_ms(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = Swin_V2_B_Weights.LANDSAT_MI_SATLAS, out_indices: list | None = None, **kwargs):
    """
    Args:
        model_bands (list[str]): A list containing the names for the bands expected by the model.
        pretrained (bool): The model is already pretrained (weights are available and can be restored) or not.
        ckpt_data (str | None): Path for a checkpoint containing the model weights.
    Returns:
        SwinEncoderWrapper
    """

    if "in_chans" not in kwargs: kwargs["in_chans"] = len(model_bands)
    model = load_model(swin_v2_b, swin_v2_b_meta, **kwargs)
    if pretrained:
        model = load_swin_weights(model, model_bands, ckpt_data, weights)
    return SwinEncoderWrapper(model, swin_v2_b_meta, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def satlas_swin_b_landsat_mi_rgb(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = Swin_V2_B_Weights.LANDSAT_SI_SATLAS, out_indices: list | None = None, **kwargs):
    """
    Args:
        model_bands (list[str]): A list containing the names for the bands expected by the model.
        pretrained (bool): The model is already pretrained (weights are available and can be restored) or not.
        ckpt_data (str | None): Path for a checkpoint containing the model weights.
    Returns:
        SwinEncoderWrapper
    """

    if "in_chans" not in kwargs: kwargs["in_chans"] = len(model_bands)
    model = load_model(swin_v2_b, swin_v2_b_meta, **kwargs)
    if pretrained:
        model = load_swin_weights(model, model_bands, ckpt_data, weights)
    return SwinEncoderWrapper(model, swin_v2_b_meta, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def satlas_swin_b_sentinel1_mi(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = Swin_V2_B_Weights.SENTINEL1_MI_SATLAS, out_indices: list | None = None, **kwargs):
    """
    Args:
        model_bands (list[str]): A list containing the names for the bands expected by the model.
        pretrained (bool): The model is already pretrained (weights are available and can be restored) or not.
        ckpt_data (str | None): Path for a checkpoint containing the model weights.
    Returns:
        SwinEncoderWrapper
    """

    if "in_chans" not in kwargs: kwargs["in_chans"] = len(model_bands)
    model = load_model(swin_v2_b, swin_v2_b_meta, **kwargs)
    if pretrained:
        model = load_swin_weights(model, model_bands, ckpt_data, weights)
    return SwinEncoderWrapper(model, swin_v2_b_meta, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def satlas_swin_b_sentinel1_si(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = Swin_V2_B_Weights.SENTINEL1_SI_SATLAS, out_indices: list | None = None, **kwargs):
    """
    Args:
        model_bands (list[str]): A list containing the names for the bands expected by the model.
        pretrained (bool): The model is already pretrained (weights are available and can be restored) or not.
        ckpt_data (str | None): Path for a checkpoint containing the model weights.
    Returns:
        SwinEncoderWrapper
    """

    if "in_chans" not in kwargs: kwargs["in_chans"] = len(model_bands)
    model = load_model(swin_v2_b, swin_v2_b_meta, **kwargs)
    if pretrained:
        model = load_swin_weights(model, model_bands, ckpt_data, weights)
    return SwinEncoderWrapper(model, swin_v2_b_meta, weights, out_indices)


def load_swin_weights(model: nn.Module, model_bands, ckpt_data: str, weights: Weights, input_size: int = 224, custom_weight_proj: str = "features.0.0.weight") -> nn.Module:
    
    pretrained_bands = get_pretrained_bands(weights.meta["bands"])
    
    if ckpt_data is not None:
        if ckpt_data.find("https://hf.co/") > -1:
            repo_id = ckpt_data.split("/resolve/")[0].replace("https://hf.co/", '')
            filename = ckpt_data.split("/")[-1]
            ckpt_data = huggingface_hub.hf_hub_download(repo_id=repo_id, filename=filename)

        checkpoint_model = torch.load(ckpt_data, map_location="cpu", weights_only=True)
        state_dict = model.state_dict()
        
        for k in ["head.weight", "head.bias"]:
                if (
                    k in checkpoint_model
                    and checkpoint_model[k].shape != state_dict[k].shape
                ):
                    logging.info(f"Removing key {k} from pretrained checkpoint")
                    del checkpoint_model[k]
    
        checkpoint_model = select_patch_embed_weights(checkpoint_model, model, pretrained_bands, model_bands, custom_weight_proj)
        # load pre-trained model
        msg = model.load_state_dict(checkpoint_model, strict=False)
    
        logging.info(msg)
    else:
        if weights is not None:
            checkpoint_model = weights.get_state_dict(progress=True)
            checkpoint_model = select_patch_embed_weights(checkpoint_model, model, pretrained_bands, model_bands, custom_weight_proj)
            missing_keys, unexpected_keys = model.load_state_dict(checkpoint_model, strict=False)
            assert set(missing_keys) <= set()
            assert not unexpected_keys

    
    return model



# Copyright contributors to the Terratorch project
from typing import List
import logging 
import torch
import torch.nn.functional as F  # noqa: N812
from segmentation_models_pytorch.base import SegmentationModel
from torch import nn

from terratorch.models.heads import RegressionHead, SegmentationHead
from terratorch.models.model import AuxiliaryHeadWithDecoderWithoutInstantiatedHead, Model, ModelOutput
from terratorch.models.utils import pad_images, get_image_size, center_crop

def freeze_module(module: nn.Module):
    for param in module.parameters():
        param.requires_grad_(False)


class PixelWiseModel(Model, SegmentationModel):
    """Model that encapsulates encoder and decoder and heads
    Expects decoder to have a "forward_features" method, an embed_dims property
    and optionally a "prepare_features_for_image_model" method.
    """

    def __init__(
        self,
        task: str,
        encoder: nn.Module,
        decoder: nn.Module,
        head_kwargs: dict,
        patch_size: int = None, 
        padding: str = None,
        decoder_includes_head: bool = False,
        auxiliary_heads: list[AuxiliaryHeadWithDecoderWithoutInstantiatedHead] | None = None,
        neck: nn.Module | None = None,
        rescale: bool = True,  # noqa: FBT002, FBT001
        image_size_out: tuple[int, int] | None = None,
    ) -> None:
        """Constructor

        Args:
            task (str): Task to be performed. One of segmentation or regression.
            encoder (nn.Module): Encoder to be used
            decoder (nn.Module): Decoder to be used
            head_kwargs (dict): Arguments to be passed at instantiation of the head.
            patch_size (int, optional): Patch size to be used during inference.
            padding (str, optional): Padding to be used during inference.
            decoder_includes_head (bool): Whether the decoder already incldes a head. If true, a head will not be added. Defaults to False.
            auxiliary_heads (list[AuxiliaryHeadWithDecoderWithoutInstantiatedHead] | None, optional): List of
                AuxiliaryHeads with heads to be instantiated. Defaults to None.
            neck (nn.Module | None): Module applied between backbone and decoder.
                Defaults to None, which applies the identity.
            rescale (bool, optional): Rescale the output of the model if it has a different size than the ground truth.
                Uses bilinear interpolation. Defaults to True.
            image_size_out (tuple[int, int] | None): The desired (Height, Width) size of the output image or mask. This is used to ensure the model produces the correct output shape.
                If set to **None** (default), the size is dynamically determined: either set by the 'image_size' property during the forward pass, or inferred directly from the size of the input image.
        """
        super().__init__()

        self.task = task
        self.encoder = encoder
        self.decoder = decoder
        self.head = (
            self._get_head(task, decoder.out_channels, head_kwargs) if not decoder_includes_head else nn.Identity()
        )

        if auxiliary_heads is not None:
            aux_heads = {}
            for aux_head_to_be_instantiated in auxiliary_heads:
                aux_head: nn.Module = self._get_head(
                    task, aux_head_to_be_instantiated.decoder.out_channels, head_kwargs
                ) if not aux_head_to_be_instantiated.decoder_includes_head else nn.Identity()
                aux_head = nn.Sequential(aux_head_to_be_instantiated.decoder, aux_head)
                aux_heads[aux_head_to_be_instantiated.name] = aux_head
        else:
            aux_heads = {}
        self.aux_heads = nn.ModuleDict(aux_heads)

        if neck is not None:
            self.neck = neck
        elif hasattr(self.encoder, "prepare_features_for_image_model"):
            # only for backwards compatibility with pre-neck times.
            def model_defined_neck(x, **kwargs):
                return self.encoder.prepare_features_for_image_model(x)  # Drop kwargs
            self.neck = model_defined_neck
        else:
            self.neck = lambda x, image_size: x

        self.rescale = rescale
        self.image_size_out = image_size_out
        self.patch_size = patch_size
        self.padding = padding

    def freeze_encoder(self):
        if hasattr(self.encoder, "freeze"):
            self.encoder.freeze()
        else:
            freeze_module(self.encoder)

    def freeze_decoder(self):
        freeze_module(self.decoder)

    def freeze_head(self):
        freeze_module(self.head)

    # TODO: do this properly
    def check_input_shape(self, x: torch.Tensor) -> bool:  # noqa: ARG002
        return True

    @staticmethod
    def _check_for_single_channel_and_squeeze(x):
        if x.shape[1] == 1:
            x = x.squeeze(1)
        return x

    def forward(self, x: torch.Tensor, **kwargs) -> ModelOutput:
        """Sequentially pass `x` through model`s encoder, decoder and heads"""

        image_size_out = (self.image_size_out
                          or kwargs.get('image_size', None)
                          or get_image_size(x))

        if self.patch_size and self.padding is not None:
            x = pad_images(x, self.patch_size, self.padding)
        input_size = get_image_size(x)

        features = self.encoder(x, **kwargs)
        features = self.neck(features, image_size=input_size)

        decoder_output = self.decoder([f.clone() for f in features])
        mask = self.head(decoder_output)

        if self.rescale and mask.shape[-2:] != input_size:
            mask = F.interpolate(mask, size=input_size, mode="bilinear")

        mask = self._check_for_single_channel_and_squeeze(mask)
        mask = center_crop(mask, image_size_out)

        aux_outputs = {}
        for name, decoder in self.aux_heads.items():
            aux_output = decoder([f.clone() for f in features])
            if self.rescale and aux_output.shape[-2:] != input_size:
                aux_output = F.interpolate(aux_output, size=input_size, mode="bilinear")
            aux_output = self._check_for_single_channel_and_squeeze(aux_output)
            aux_output = center_crop(aux_output, image_size_out)
            aux_outputs[name] = aux_output


        return ModelOutput(output=mask, auxiliary_heads=aux_outputs)

    def _get_head(self, task: str, input_embed_dim: int, head_kwargs):
        if task == "segmentation":
            if "num_classes" not in head_kwargs:
                msg = "num_classes must be defined for segmentation task"
                raise Exception(msg)
            return SegmentationHead(input_embed_dim, **head_kwargs)
        if task == "regression":
            return RegressionHead(input_embed_dim, **head_kwargs)
        msg = "Task must be one of segmentation or regression."
        raise Exception(msg)


import torch
import torch.utils.checkpoint as cp
from timm.layers import DropPath
from timm.models.vision_transformer import Block
from torch import nn

from terratorch.models.backbones.detr_ops.modules import MSDeformAttn


def get_reference_points(spatial_shapes: list[tuple[int, int]], device: torch.device) -> torch.Tensor:
    reference_points_list = []
    for lvl, (H_, W_) in enumerate(spatial_shapes):  # noqa: N806
        ref_y, ref_x = torch.meshgrid(
            torch.linspace(0.5, H_ - 0.5, H_, dtype=torch.float32, device=device),
            torch.linspace(0.5, W_ - 0.5, W_, dtype=torch.float32, device=device),
            indexing="ij",  # suppress warning
        )
        ref_y = ref_y.reshape(-1)[None] / H_
        ref_x = ref_x.reshape(-1)[None] / W_
        ref = torch.stack((ref_x, ref_y), -1)
        reference_points_list.append(ref)
    reference_points = torch.cat(reference_points_list, 1)
    reference_points = reference_points[:, :, None]
    return reference_points


def deform_inputs(x: torch.Tensor) -> tuple[list[torch.Tensor], list[torch.Tensor]]:
    bs, c, h, w = x.shape
    spatial_shapes = torch.as_tensor(
        [(h // 8, w // 8), (h // 16, w // 16), (h // 32, w // 32)],
        dtype=torch.long,
        device=x.device,
    )
    level_start_index = torch.cat((spatial_shapes.new_zeros((1,)), spatial_shapes.prod(1).cumsum(0)[:-1]))
    reference_points = get_reference_points([(h // 16, w // 16)], x.device)
    deform_inputs1 = [reference_points, spatial_shapes, level_start_index]

    spatial_shapes = torch.as_tensor([(h // 16, w // 16)], dtype=torch.long, device=x.device)
    level_start_index = torch.cat((spatial_shapes.new_zeros((1,)), spatial_shapes.prod(1).cumsum(0)[:-1]))
    reference_points = get_reference_points([(h // 8, w // 8), (h // 16, w // 16), (h // 32, w // 32)], x.device)
    deform_inputs2 = [reference_points, spatial_shapes, level_start_index]

    return deform_inputs1, deform_inputs2


class SpatialPriorModule(nn.Module):
    def __init__(self, inplanes: int, embed_dim: int, in_channels: int, with_cp: bool = False):
        super().__init__()
        self.with_cp = with_cp

        self.stem = nn.Sequential(
            *[
                nn.Conv2d(
                    in_channels,
                    inplanes,
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    bias=False,
                ),
                nn.SyncBatchNorm(inplanes),
                nn.ReLU(inplace=True),
                nn.Conv2d(inplanes, inplanes, kernel_size=3, stride=1, padding=1, bias=False),
                nn.SyncBatchNorm(inplanes),
                nn.ReLU(inplace=True),
                nn.Conv2d(inplanes, inplanes, kernel_size=3, stride=1, padding=1, bias=False),
                nn.SyncBatchNorm(inplanes),
                nn.ReLU(inplace=True),
                nn.MaxPool2d(kernel_size=3, stride=2, padding=1),
            ]
        )
        self.conv2 = nn.Sequential(
            *[
                nn.Conv2d(
                    inplanes,
                    2 * inplanes,
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    bias=False,
                ),
                nn.SyncBatchNorm(2 * inplanes),
                nn.ReLU(inplace=True),
            ]
        )
        self.conv3 = nn.Sequential(
            *[
                nn.Conv2d(
                    2 * inplanes,
                    4 * inplanes,
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    bias=False,
                ),
                nn.SyncBatchNorm(4 * inplanes),
                nn.ReLU(inplace=True),
            ]
        )
        self.conv4 = nn.Sequential(
            *[
                nn.Conv2d(
                    4 * inplanes,
                    4 * inplanes,
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    bias=False,
                ),
                nn.SyncBatchNorm(4 * inplanes),
                nn.ReLU(inplace=True),
            ]
        )
        self.fc1 = nn.Conv2d(inplanes, embed_dim, kernel_size=1, stride=1, padding=0, bias=True)
        self.fc2 = nn.Conv2d(2 * inplanes, embed_dim, kernel_size=1, stride=1, padding=0, bias=True)
        self.fc3 = nn.Conv2d(4 * inplanes, embed_dim, kernel_size=1, stride=1, padding=0, bias=True)
        self.fc4 = nn.Conv2d(4 * inplanes, embed_dim, kernel_size=1, stride=1, padding=0, bias=True)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        def _inner_forward(
            x: torch.Tensor,
        ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
            c1 = self.stem(x)
            c2 = self.conv2(c1)
            c3 = self.conv3(c2)
            c4 = self.conv4(c3)
            c1 = self.fc1(c1)
            c2 = self.fc2(c2)
            c3 = self.fc3(c3)
            c4 = self.fc4(c4)

            bs, dim, _, _ = c1.shape
            # c1 = c1.view(bs, dim, -1).transpose(1, 2)  # 4s
            c2 = c2.view(bs, dim, -1).transpose(1, 2)  # 8s
            c3 = c3.view(bs, dim, -1).transpose(1, 2)  # 16s
            c4 = c4.view(bs, dim, -1).transpose(1, 2)  # 32s

            return c1, c2, c3, c4

        if self.with_cp and x.requires_grad:
            outs: tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor] = cp.checkpoint(_inner_forward, x)
        else:
            outs = _inner_forward(x)
        return outs


class ConvFFN(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int | None,
        out_features: int | None,
        act_layer: type[nn.Module],
        drop: float,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.dwconv = DWConv(hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x: torch.Tensor, H: int, W: int) -> torch.Tensor:  # noqa: N803
        x = self.fc1(x)
        x = self.dwconv(x, H, W)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class DWConv(nn.Module):
    def __init__(self, dim: int):
        super().__init__()
        self.dwconv = nn.Conv2d(dim, dim, 3, 1, 1, bias=True, groups=dim)

    def forward(self, x: torch.Tensor, H: int, W: int) -> torch.Tensor:  # noqa: N803
        B, N, C = x.shape  # noqa: N806
        n = N // 21
        x1 = x[:, 0 : 16 * n, :].transpose(1, 2).view(B, C, H * 2, W * 2).contiguous()
        x2 = x[:, 16 * n : 20 * n, :].transpose(1, 2).view(B, C, H, W).contiguous()
        x3 = x[:, 20 * n :, :].transpose(1, 2).view(B, C, H // 2, W // 2).contiguous()
        x1 = self.dwconv(x1).flatten(2).transpose(1, 2)
        x2 = self.dwconv(x2).flatten(2).transpose(1, 2)
        x3 = self.dwconv(x3).flatten(2).transpose(1, 2)
        x = torch.cat([x1, x2, x3], dim=1)
        return x


class Extractor(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        n_points: int,
        n_levels: int,
        deform_ratio: float,
        with_cffn: bool,
        cffn_ratio: float,
        drop: float,
        drop_path: float,
        norm_layer: type[nn.Module],
        with_cp: bool,
    ):
        super().__init__()
        self.query_norm = norm_layer(dim)
        self.feat_norm = norm_layer(dim)
        self.attn = MSDeformAttn(
            d_model=dim,
            n_levels=n_levels,
            n_heads=num_heads,
            n_points=n_points,
            ratio=deform_ratio,
        )
        self.with_cffn = with_cffn
        self.with_cp = with_cp
        if with_cffn:
            self.ffn = ConvFFN(
                in_features=dim,
                hidden_features=int(dim * cffn_ratio),
                drop=drop,
                out_features=None,
                act_layer=nn.GELU,
            )
            self.ffn_norm = norm_layer(dim)
            self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

    def forward(
        self,
        query: torch.Tensor,
        reference_points: torch.Tensor,
        feat: torch.Tensor,
        spatial_shapes: torch.Tensor,
        level_start_index: torch.Tensor,
        H: int,  # noqa: N803
        W: int,  # noqa: N803
    ) -> torch.Tensor:
        def _inner_forward(query: torch.Tensor, feat: torch.Tensor) -> torch.Tensor:
            attn = self.attn(
                self.query_norm(query),
                reference_points,
                self.feat_norm(feat),
                spatial_shapes,
                level_start_index,
                None,
            )
            query = query + attn

            if self.with_cffn:
                query = query + self.drop_path(self.ffn(self.ffn_norm(query), H, W))
            return query

        if self.with_cp and query.requires_grad:
            query = cp.checkpoint(_inner_forward, query, feat)
        else:
            query = _inner_forward(query, feat)

        return query


class Injector(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        n_points: int,
        n_levels: int,
        deform_ratio: float,
        norm_layer: type[nn.Module],
        init_values: float,
        with_cp: bool,
    ):
        super().__init__()
        self.with_cp = with_cp
        self.query_norm = norm_layer(dim)
        self.feat_norm = norm_layer(dim)
        self.attn = MSDeformAttn(
            d_model=dim,
            n_levels=n_levels,
            n_heads=num_heads,
            n_points=n_points,
            ratio=deform_ratio,
        )
        self.gamma = nn.Parameter(init_values * torch.ones(dim), requires_grad=True)

    def forward(
        self,
        query: torch.Tensor,
        reference_points: torch.Tensor,
        feat: torch.Tensor,
        spatial_shapes: torch.Tensor,
        level_start_index: torch.Tensor,
    ) -> torch.Tensor:
        def _inner_forward(query: torch.Tensor, feat: torch.Tensor) -> torch.Tensor:
            attn: torch.Tensor = self.attn(
                self.query_norm(query),
                reference_points,
                self.feat_norm(feat),
                spatial_shapes,
                level_start_index,
                None,
            )
            return query + self.gamma * attn

        if self.with_cp and query.requires_grad:
            query = cp.checkpoint(_inner_forward, query, feat)
        else:
            query = _inner_forward(query, feat)

        return query


class InteractionBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        n_points: int,
        norm_layer: type[nn.Module],
        drop: float,
        drop_path: float,
        with_cffn: bool,
        cffn_ratio: float,
        init_values: float,
        deform_ratio: float,
        extra_extractor: bool,
        with_cp: bool,
    ):
        super().__init__()

        self.injector = Injector(
            dim=dim,
            n_levels=3,
            num_heads=num_heads,
            init_values=init_values,
            n_points=n_points,
            norm_layer=norm_layer,
            deform_ratio=deform_ratio,
            with_cp=with_cp,
        )
        self.extractor = Extractor(
            dim=dim,
            n_levels=1,
            num_heads=num_heads,
            n_points=n_points,
            norm_layer=norm_layer,
            deform_ratio=deform_ratio,
            with_cffn=with_cffn,
            cffn_ratio=cffn_ratio,
            drop=drop,
            drop_path=drop_path,
            with_cp=with_cp,
        )
        self.extra_extractors: nn.Sequential | None
        if extra_extractor:
            self.extra_extractors = nn.Sequential(
                *[
                    Extractor(
                        dim=dim,
                        num_heads=num_heads,
                        n_points=n_points,
                        norm_layer=norm_layer,
                        with_cffn=with_cffn,
                        cffn_ratio=cffn_ratio,
                        deform_ratio=deform_ratio,
                        drop=drop,
                        drop_path=drop_path,
                        with_cp=with_cp,
                        n_levels=1,
                    )
                    for _ in range(2)
                ]
            )
        else:
            self.extra_extractors = None

    def forward(
        self,
        x: torch.Tensor,
        c: torch.Tensor,
        blocks: list[Block],
        deform_inputs1: list[torch.Tensor],
        deform_inputs2: list[torch.Tensor],
        H: int,  # noqa: N803
        W: int,  # noqa: N803
    ) -> tuple[torch.Tensor, torch.Tensor]:
        x = self.injector(
            query=x,
            reference_points=deform_inputs1[0],
            feat=c,
            spatial_shapes=deform_inputs1[1],
            level_start_index=deform_inputs1[2],
        )
        for idx, blk in enumerate(blocks):
            # x = blk(x, H, W)
            x = blk(x)  # we don't need H and W as we don't use windowed attn for now
        c = self.extractor(
            query=c,
            reference_points=deform_inputs2[0],
            feat=x,
            spatial_shapes=deform_inputs2[1],
            level_start_index=deform_inputs2[2],
            H=H,
            W=W,
        )
        if self.extra_extractors is not None:
            for extractor in self.extra_extractors:
                c = extractor(
                    query=c,
                    reference_points=deform_inputs2[0],
                    feat=x,
                    spatial_shapes=deform_inputs2[1],
                    level_start_index=deform_inputs2[2],
                    H=H,
                    W=W,
                )
        return x, c



# Copyright contributors to the Terratorch project

from dataclasses import dataclass
from torch import nn
import pdb
from terratorch.models.model import (
    Model,
    ModelFactory,
)
from terratorch.models.necks import build_neck_list, NeckSequential
from terratorch.models.model import ModelOutput
from terratorch.models.utils import extract_prefix_keys
from terratorch.registry import MODEL_FACTORY_REGISTRY

import torchvision.models.detection
from torchvision.models.detection.retinanet import RetinaNetHead
from torchvision.models.detection.rpn import AnchorGenerator
from torchvision.ops import MultiScaleRoIAlign

import numpy as np
from functools import partial
import torch
import pdb
from .utils import _get_backbone, TerratorchGeneralizedRCNNTransform

SUPPORTED_TASKS = ['object_detection']


def _check_all_args_used(kwargs):
    """
    Check if all arguments are used.

    Args:
        kwargs: dict: Dictionary of arguments.
    
    """
    if kwargs:
        msg = f"arguments {kwargs} were passed but not used."
        raise ValueError(msg)


@MODEL_FACTORY_REGISTRY.register
class ObjectDetectionModelFactory(ModelFactory):
    def build_model(
        self,
        task: str,
        backbone: str | nn.Module,
        framework: str,
        num_classes: int | None = None,
        necks: list[dict] | None = None,
        **kwargs,
    ) -> Model:
        """
        Generic model factory that combines an encoder and necks with the detection models, called framework, in torchvision.detection.

        Further arguments to be passed to the backbone_ and framework_.

        Args:
            task (str): Task to be performed. Currently supports "object_detection".
            backbone (str, nn.Module): Backbone to be used. If a string, will look for such models in the different
                registries supported (internal terratorch registry, timm, ...). If a torch nn.Module, will use it
                directly. The backbone should have and `out_channels` attribute and its `forward` should return a list[Tensor].
            framework (str): object detection framework to be used between "faster-rcnn", "fcos", "retinanet" for object detection and "mask-rcnn" for instance segmentation.
            num_classes (int, optional): Number of classes. None for regression tasks.
            necks (list[dict]): nn.Modules to be called in succession on encoder features
                before passing them to the decoder. Should be registered in the NECKS_REGISTRY registry.
                Expects each one to have a key "name" and subsequent keys for arguments, if any.
                Defaults to None, which applies the identity function.

        Returns:
            nn.Module: Full torchvision detection model.
        """
        task = task.lower()
        if task not in SUPPORTED_TASKS:
            msg = f"Task {task} not supported. Please choose one of {SUPPORTED_TASKS}"
            raise NotImplementedError(msg)
        backbone_kwargs, kwargs = extract_prefix_keys(kwargs, "backbone_")
        framework_kwargs, kwargs = extract_prefix_keys(kwargs, "framework_")
        
        backbone = _get_backbone(backbone, **backbone_kwargs)
        if 'in_channels' in kwargs.keys():
            in_channels = kwargs['in_channels']
        else:
            in_channels = len(backbone_kwargs["model_bands"]) if "model_bands" in backbone_kwargs.keys() else len(backbone_kwargs["bands"])

        try:
            out_channels = backbone.out_channels
        except AttributeError as e:
            msg = "backbone must have out_channels attribute"
            raise AttributeError(msg) from e
        # pdb.set_trace()
        if necks is None:
            necks = []
        neck_list, channel_list = build_neck_list(necks, out_channels)

        neck_module = NeckSequential(*neck_list)

        combined_backbone = BackboneWrapper(backbone, neck_module, channel_list)
        # pdb.set_trace()
        
        if framework == 'faster-rcnn':

            sizes = ((32), (64), (128), (256), (512))
            sizes = sizes[:len(combined_backbone.channel_list)]
            aspect_ratios = ((0.5, 1.0, 2.0),) * len(sizes)
            anchor_generator = AnchorGenerator(sizes=sizes, aspect_ratios=aspect_ratios)

            roi_pooler = MultiScaleRoIAlign(
                featmap_names=['feat0', 'feat1', 'feat2', 'feat3'], output_size=7, sampling_ratio=2
            )

            model = torchvision.models.detection.FasterRCNN(
                combined_backbone,
                num_classes,
                rpn_anchor_generator=anchor_generator,
                box_roi_pool=roi_pooler,
                _skip_resize=True,
                image_mean = np.repeat(0, in_channels),
                image_std = np.repeat(1, in_channels),
                **framework_kwargs
            )
            
        elif framework == 'fcos':

            sizes = ((8,), (16,), (32,), (64,), (128,), (256,))
            sizes=sizes[:len(combined_backbone.channel_list)]
            aspect_ratios = ((1.0,), (1.0,), (1.0,), (1.0,), (1.0,), (1.0,)) * len(sizes)
            anchor_generator = AnchorGenerator(
                sizes=sizes,
                aspect_ratios=aspect_ratios,
            )

            model = torchvision.models.detection.FCOS(
                combined_backbone, 
                num_classes,
                anchor_generator=anchor_generator, 
                _skip_resize=True,
                image_mean = np.repeat(0, in_channels),
                image_std = np.repeat(1, in_channels),
                **framework_kwargs

            )
        elif framework == 'retinanet':

            sizes = (
                (16, 20, 25),
                (32, 40, 50),
                (64, 80, 101),
                (128, 161, 203),
                (256, 322, 406),
                (512, 645, 812),
            )
            sizes=sizes[:len(combined_backbone.channel_list)]
            aspect_ratios = ((0.5, 1.0, 2.0),) * len(sizes)
            anchor_generator = AnchorGenerator(sizes, aspect_ratios)
            head = RetinaNetHead(
                combined_backbone.out_channels,
                anchor_generator.num_anchors_per_location()[0],
                num_classes,
                norm_layer=partial(torch.nn.GroupNorm, 32),
            )

            model = torchvision.models.detection.RetinaNet(
                combined_backbone,
                num_classes,
                anchor_generator=anchor_generator,
                head=head,
                _skip_resize=True,
                image_mean=np.repeat(0, in_channels),
                image_std=np.repeat(1, in_channels),
                **framework_kwargs
            )

        elif framework == 'mask-rcnn':

            sizes = ((32), (64), (128), (256), (512))
            sizes = sizes[:len(combined_backbone.channel_list)]
            aspect_ratios = ((0.5, 1.0, 2.0),) * len(sizes)
            anchor_generator = AnchorGenerator(sizes=sizes, aspect_ratios=aspect_ratios)

            rpn_head = torchvision.models.detection.faster_rcnn.RPNHead(combined_backbone.out_channels, anchor_generator.num_anchors_per_location()[0], conv_depth=2)
            box_head = torchvision.models.detection.faster_rcnn.FastRCNNConvFCHead(
                (combined_backbone.out_channels, 7, 7), [256, 256, 256, 256], [1024], norm_layer=nn.BatchNorm2d
            )
            mask_head = torchvision.models.detection.mask_rcnn.MaskRCNNHeads(combined_backbone.out_channels, [256, 256, 256, 256], 1, norm_layer=nn.BatchNorm2d)
            roi_pooler = MultiScaleRoIAlign(
                featmap_names=['feat0', 'feat1', 'feat2', 'feat3'], output_size=7, sampling_ratio=2
            )

            model = torchvision.models.detection.MaskRCNN(
                combined_backbone,
                num_classes=num_classes,
                rpn_anchor_generator=anchor_generator,
                rpn_head=rpn_head,
                box_head=box_head,
                box_roi_pool=roi_pooler,
                mask_roi_pool=roi_pooler,
                mask_head=mask_head,
                _skip_resize=True,
                image_mean=np.repeat(0, in_channels),
                image_std=np.repeat(1, in_channels),
                **framework_kwargs
            )

        else:
            raise ValueError(f"Framework type '{framework}' is not valid.")
        # some decoders already include a head
        # for these, we pass the num_classes to them
        # others dont include a head
        # for those, we dont pass num_classes
        # model.transform = IdentityTransform()
        model.transform = TerratorchGeneralizedRCNNTransform(model.transform.min_size, 
                                                             model.transform.max_size, 
                                                             model.transform.image_mean, 
                                                             model.transform.image_std,  
                                                             model.transform.size_divisible, 
                                                             model.transform.fixed_size,_skip_resize=model.transform._skip_resize)

        return ObjectDetectionModel(model, framework)


class BackboneWrapper(nn.Module):

    def __init__(self, backbone, necks, channel_list):
        """
        BackboneWrapper class that wraps a backbone and necks.

        Args:
            backbone (nn.Module): Backbone module.
            necks (nn.Module): Necks module.
            channel_list (list): List of output channels for each neck.

        Returns:
            dict: Dictionary of output features from necks.
        """   
        super().__init__()
        self.backbone = backbone
        self.necks = necks
        self.out_channels = channel_list[-1]
        self.channel_list = channel_list

    def forward(self, x, **kwargs):
        """
        Forward pass of the model.
        
        Parameters:
        x (torch.Tensor): Input tensor.
        **kwargs (dict): Additional keyword arguments.
        
        Returns:
        torch.Tensor: Output tensor.
        """
        x = self.backbone(x, **kwargs)
        x = self.necks(x)
        
        return x


class ObjectDetectionModel(Model):
    def __init__(self, torchvision_model, model_name) -> None:

        """
        Wrapper for torchvision models.

        Args:
            torchvision_model (torchvision.models): A torchvision model.
            model_name (str): The name of the model.

        Returns:
            None
        """

        super().__init__()
        self.torchvision_model = torchvision_model
        self.model_name = model_name

    def forward(self, x, *args, **kwargs):
        """
        Forward pass of the model.
        
        Parameters:
        x (torch.Tensor): Input tensor.
        **kwargs (dict): Additional keyword arguments.
        
        Returns:
        torch.Tensor: Output tensor.
        """
        return ModelOutputObjectDetection(self.torchvision_model(x, *args, **kwargs))

    def freeze_encoder(self):
        """
        Freeze the encoder of the model.
        """
        for param in self.torchvision_model.backbone.parameters():
            param.requires_grad = False

    def freeze_decoder(self):
        """
        Freeze the decoder of the model.
        """
        if self.model_name == 'faster-rcnn':
            for param in self.torchvision_model.rpn.parameters():
                param.requires_grad = False
                for param in self.torchvision_model.roi_heads.parameters():
                    param.requires_grad = False
        elif self.model_name == 'fcos':
            for param in self.torchvision_model.head.parameters():
                param.requires_grad = False
        elif self.model_name == 'retinanet':
            for param in self.torchvision_model.head.parameters():
                param.requires_grad = False
        else:
            raise ValueError(f"Model type '{self.model_name}' is not valid.")


@dataclass
class ModelOutputObjectDetection(ModelOutput):
    output: dict


class ImageContainer:
    def __init__(self, tensor):
        self.tensors = tensor




import warnings
import re
import torch
from torch import nn
from tokenizers import Tokenizer, AddedToken
from tokenizers.models import WordPiece
from tokenizers.pre_tokenizers import BertPreTokenizer
from tokenizers.normalizers import BertNormalizer
from tokenizers.decoders import WordPiece as WordPieceDecoder


def build_blank_wordpiece():
    # Minimal tokens needed for testing
    vocab = {"[PAD]": 0, "[UNK]": 1, "[SOS]": 2,"[EOS]": 3, "[S_0]": 4, "[S_1]": 5, "[S_2]": 6}
    added_tokens = [AddedToken(w, normalized=False, special=True) for w in vocab.keys()]
    tok = Tokenizer(WordPiece(vocab, unk_token="[UNK]"))
    tok.normalizer = BertNormalizer()
    tok.pre_tokenizer = BertPreTokenizer()
    tok.decoder = WordPieceDecoder()
    tok.add_tokens(added_tokens)
    return tok


def capitalize_sentences(text):
    # Split text into sentences using a regex that looks for sentence end punctuation
    sentences = re.split('([.!?] *)', text)
    capitalized = ''.join([s.capitalize() for s in sentences])
    return capitalized


class CaptionTokenizer(nn.Module):
    def __init__(self, tokenizer_file, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if tokenizer_file is not None:
            self.text_tokenizer = Tokenizer.from_file(tokenizer_file)
        else:
            self.text_tokenizer = build_blank_wordpiece()  # un-trained
        self.text_tokenizer.enable_padding()

    def encode(self, text: list[str], device: torch.device, eos_id=3, *args, **kwargs) -> dict[str, torch.Tensor]:
        """
        Args:
            text list[str]: Text to be tokenized
            device: torch.device
        Returns:
            dict for generation sampler input
        """
        # Add start token
        text = [t + " [S_1]" for t in text]

        # Tokenize
        tok_ids = [t.ids for t in self.text_tokenizer.encode_batch(text, add_special_tokens=True)]

        # Add EOS token
        tok_ids = [t + [eos_id] for t in tok_ids]


        tok_ids = torch.tensor(tok_ids, device=device)

        text_dict = {
            "tensor": tok_ids,
            "input_mask": torch.zeros_like(tok_ids, dtype=torch.bool, device=device),
            "target_mask": torch.ones_like(tok_ids, dtype=torch.bool, device=device),
            "decoder_attention_mask": torch.zeros_like(tok_ids, dtype=torch.bool, device=device),
        }

        return text_dict

    def decode_text(self, mod_dict, key="caption"):
        """
        Decodes a text sequence from a model dictionary.

        Args:
            mod_dict (dict): Model output dictionary.
            key (str): Key of the text modality to decode.
        """
        decoded_texts = []

        for i in range(mod_dict[key]["tensor"].shape[0]):
            seq = mod_dict[key]["tensor"][i]
            seq = seq[mod_dict[key]["input_mask"][i] == 0]
            seq = seq.tolist()

            merged_text = self.text_tokenizer.decode(seq, skip_special_tokens=False)

            decoded_texts.append(merged_text.replace(" [EOS]", ""))

        decoded_texts = list(map(capitalize_sentences, decoded_texts))

        return decoded_texts


class CoordsTokenizer(nn.Module):
    def __init__(self, tokenizer_file, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if tokenizer_file is not None:
            self.text_tokenizer = Tokenizer.from_file(tokenizer_file)
        else:
            self.text_tokenizer = build_blank_wordpiece()  # un-trained

    def encode(self, coords: torch.Tensor, *args, **kwargs) -> dict[str, torch.Tensor]:
        """
        Encodes coords to token ids. Returns tuple to be compatible with image tokenizers.

        Args:
            coords (torch.Tensor): Center coordinates of image with shape [B, 2] with [lon, lat] values in second dim.

        Returns:
            tok_ids(tuple[torch.Tensor]): Token ids with shape [B, 2]
        """
        if coords.shape[1] != 2:
            raise ValueError(f"Expect coords data in shape [batch, 2] with [lon, lat] values, "
                             f"got coords with shape {coords.shape}.")

        # Align coords with 0.25 degree grid
        coords = (coords * 4).round() / 4
        device = coords.device

        coords = [f"lat={c[1].item():.2f} lon={c[0].item():.2f} [EOS]" for c in coords]

        # Tokenize
        tok_ids = [t.ids for t in self.text_tokenizer.encode_batch(coords, add_special_tokens=True)]
        tok_ids = torch.tensor(tok_ids, device=device)

        coords_dict = {
            "tensor": tok_ids,
            "input_mask": torch.zeros_like(tok_ids, dtype=torch.bool, device=device),
            "target_mask": torch.ones_like(tok_ids, dtype=torch.bool, device=device),
            "decoder_attention_mask": torch.zeros_like(tok_ids, dtype=torch.bool, device=device),
        }

        return coords_dict

    def decode_text(self, mod_dict, key="coords"):
        """
        Decodes a coordinate sequence from a modality dictionary.

        Args:
            mod_dict (dict): Model output dictionary.
            key (str): Key of the coords modality to decode.
        """
        coords = []

        B = mod_dict[key]["tensor"].shape[0]

        for i in range(B):
            seq = mod_dict[key]["tensor"][i].tolist()[:2]

            text = self.text_tokenizer.decode(seq, skip_special_tokens=False)

            try:
                lat, lon = text.split(" ")
                coords.append([float(lon.strip("lon=")), float(lat.strip("lat="))])
            except Exception as e:
                warnings.warn(f"Coordinate generation did not work correctly, generated text: {text} (Error: {e}). "
                              f"Returning NaN.")
                coords.append([torch.nan, torch.nan])

        return coords



import copy
import gc

import pytest
import torch
from torchvision.models.detection.image_list import ImageList

# Adjust the import below to match where your utils.py lives in your repo.
# If utils.py sits at terratorch/datamodules/utils.py, this import is correct:
from terratorch.models.utils import TerratorchGeneralizedRCNNTransform


@pytest.fixture
def dummy_images():
    # Two images with different spatial sizes but same channels
    # C, H, W

    return [torch.randn(3, 32, 48), torch.randn(3, 32, 48)]


@pytest.fixture
def dummy_targets():
    # Two simple target dicts (common detection fields)
    t1 = {"boxes": torch.tensor([[1.0, 2.0, 10.0, 20.0]]), "labels": torch.tensor([1])}
    t2 = {"boxes": torch.tensor([[5.0, 5.0, 12.0, 12.0]]), "labels": torch.tensor([2])}
    return [t1, t2]


def _make_transform():
    """
    NOTE:
      In your implementation, the class defines `init(...)` (not `__init__`).
      That means the parent `GeneralizedRCNNTransform.__init__` is used.
      So we must pass the parent ctor args here.
    """
    # Typical values used by torchvision's GeneralizedRCNNTransform
    # min_size and max_size can be single ints; stats are per-channel (3)
    return TerratorchGeneralizedRCNNTransform(
        min_size=800,
        max_size=1333,
        image_mean=[0.0, 0.0, 0.0],
        image_std=[1.0, 1.0, 1.0],
        size_divisible=32,
        fixed_size=None,
    )


def test_forward_returns_imagelist_and_targets(dummy_images, dummy_targets):
    tfm = _make_transform()
    image_list, targets_out = tfm.forward(dummy_images, dummy_targets)

    # Types
    assert isinstance(image_list, ImageList)
    assert isinstance(targets_out, list)
    assert len(targets_out) == len(dummy_targets)

    # Batched tensor of shape (N, C, H, W); since forward stacks as-is,
    # H and W must match the per-image sizes for each sample in image_list.image_sizes
    assert isinstance(image_list.tensors, torch.Tensor)
    assert image_list.tensors.ndim == 4
    assert image_list.tensors.shape[0] == 2  # batch size N

    # image_sizes preserved and are List[Tuple[int, int]]
    assert image_list.image_sizes == [(32, 48), (32, 48)]

    # Targets are shallow-copied dicts: new dict objects, same tensor objects
    for t_in, t_out in zip(dummy_targets, targets_out, strict=False):
        assert t_in is not t_out  # dict was copied
        for k in t_in:
            # tensors are same objects (shallow copy), not deep-copied
            assert t_in[k] is t_out[k]
            assert torch.equal(t_in[k], t_out[k])

    gc.collect()


def test_forward_without_targets(dummy_images):
    tfm = _make_transform()
    image_list, targets_out = tfm.forward(dummy_images, targets=None)

    assert isinstance(image_list, ImageList)
    assert targets_out is None
    assert image_list.image_sizes == [(32, 48), (32, 48)]
    # Check batched tensor first dimension equals number of images
    assert image_list.tensors.shape[0] == len(dummy_images)
    gc.collect()


def test_images_are_copied_not_referenced(dummy_images, dummy_targets):
    """
    The implementation constructs a new batched tensor.
    Mutating the original list should not change ImageList internals.
    """
    tfm = _make_transform()
    image_list, _ = tfm.forward(dummy_images, dummy_targets)

    # Keep an old reference
    old_batched = image_list.tensors.clone()

    # Mutate original images after forward
    dummy_images[0].add_(10.0)
    dummy_images[1].mul_(0.0)

    # ImageList should be unchanged
    assert torch.equal(image_list.tensors, old_batched)
    gc.collect()


def test_invalid_image_rank_raises():
    """
    Code asserts that the last two elements of img.shape are H and W.
    If we pass a wrongly shaped tensor (e.g., [C, H] or [H, W] without C),
    the len(image_size) check will fail and raise AssertionError.
    """
    tfm = _make_transform()
    bad_img = torch.randn(3, 32, 32)  # 2D tensor (no channel dim)
    ok_img = torch.randn(3, 16, 16)
    with pytest.raises((AssertionError, RuntimeError)):
        tfm.forward([bad_img, ok_img], targets=None)
    gc.collect()


def test_targets_are_shallow_copied(dummy_images, dummy_targets):
    """
    Verify targets are shallow-copied dicts: outer dict copied, tensors shared.
    """
    tfm = _make_transform()
    targets_in = dummy_targets
    image_list, targets_out = tfm.forward(dummy_images, targets=targets_in)

    # Different dict objects
    for t_in, t_out in zip(targets_in, targets_out, strict=False):
        assert t_in is not t_out
        # Modifying the returned dict doesn't mutate original dict object
        t_out["new_key"] = torch.tensor([1])
        assert "new_key" not in t_in

        # But tensors for existing keys are the same objects (shallow copy)
        assert (
            (t_in["boxes"] is targets_out[0]["boxes"])
            or (t_in["labels"] is targets_out[1]["labels"])
            or (t_in["boxes"] is targets_out[1]["boxes"])
            or (t_in["labels"] is targets_out[0]["labels"])
        )

    gc.collect()
